import tempfile
from pathlib import Path
from random import randint

import pytest

import gitlab


TEMP_DIR = tempfile.gettempdir()


def random_id():
    """
    Helper to ensure new resource creation does not clash with
    existing resources, for example when a previous test deleted a
    resource but GitLab is still deleting it asynchronously in the
    background. TODO: Expand to make it 100% safe.
    """
    return randint(9, 9999)


@pytest.fixture(scope="session")
def CONFIG():
    return Path(TEMP_DIR) / "python-gitlab.cfg"


@pytest.fixture(scope="session")
def gl(CONFIG):
    """Helper instance to make fixtures and asserts directly via the API."""
    return gitlab.Gitlab.from_config("local", [CONFIG])


@pytest.fixture(scope="module")
def group(gl):
    """Group fixture for group API resource tests."""
    _id = random_id()
    data = {
        "name": f"test-group-{_id}",
        "path": f"group-{_id}",
    }
    group = gl.groups.create(data)

    yield group

    try:
        group.delete()
    except gitlab.exceptions.GitlabDeleteError as e:
        print(f"Group already deleted: {e}")


@pytest.fixture(scope="module")
def project(gl):
    """Project fixture for project API resource tests."""
    _id = random_id()
    name = f"test-project-{_id}"

    project = gl.projects.create(name=name)

    yield project

    try:
        project.delete()
    except gitlab.exceptions.GitlabDeleteError as e:
        print(f"Project already deleted: {e}")


@pytest.fixture(scope="module")
def user(gl):
    """User fixture for user API resource tests."""
    _id = random_id()
    email = f"user{_id}@email.com"
    username = f"user{_id}"
    name = f"User {_id}"
    password = "fakepassword"

    user = gl.users.create(email=email, username=username, name=name, password=password)

    yield user

    try:
        user.delete()
    except gitlab.exceptions.GitlabDeleteError as e:
        print(f"User already deleted: {e}")


@pytest.fixture(scope="module")
def issue(project):
    """Issue fixture for issue API resource tests."""
    _id = random_id()
    data = {"title": f"Issue {_id}", "description": f"Issue {_id} description"}

    return project.issues.create(data)


@pytest.fixture(scope="module")
def label(project):
    """Label fixture for project label API resource tests."""
    _id = random_id()
    data = {
        "name": f"prjlabel{_id}",
        "description": f"prjlabel1 {_id} description",
        "color": "#112233",
    }

    return project.labels.create(data)


@pytest.fixture(scope="module")
def group_label(group):
    """Label fixture for group label API resource tests."""
    _id = random_id()
    data = {
        "name": f"grplabel{_id}",
        "description": f"grplabel1 {_id} description",
        "color": "#112233",
    }

    return group.labels.create(data)


@pytest.fixture(scope="module")
def variable(project):
    """Variable fixture for project variable API resource tests."""
    _id = random_id()
    data = {"key": f"var{_id}", "value": f"Variable {_id}"}

    return project.variables.create(data)


@pytest.fixture(scope="module")
def deploy_token(project):
    """Deploy token fixture for project deploy token API resource tests."""
    _id = random_id()
    data = {
        "name": f"token-{_id}",
        "username": "root",
        "expires_at": "2021-09-09",
        "scopes": "read_registry",
    }

    return project.deploytokens.create(data)


@pytest.fixture(scope="module")
def group_deploy_token(group):
    """Deploy token fixture for group deploy token API resource tests."""
    _id = random_id()
    data = {
        "name": f"group-token-{_id}",
        "username": "root",
        "expires_at": "2021-09-09",
        "scopes": "read_registry",
    }

    return group.deploytokens.create(data)



"""
GitLab API:
https://docs.gitlab.com/ee/api/instance_level_ci_variables.html
https://docs.gitlab.com/ee/api/project_level_variables.html
https://docs.gitlab.com/ee/api/group_level_variables.html
"""


def test_instance_variables(gl):
    variable = gl.variables.create({"key": "key1", "value": "value1"})
    assert variable.value == "value1"
    assert len(gl.variables.list()) == 1

    variable.value = "new_value1"
    variable.save()
    variable = gl.variables.get(variable.key)
    assert variable.value == "new_value1"

    variable.delete()
    assert len(gl.variables.list()) == 0


def test_group_variables(group):
    variable = group.variables.create({"key": "key1", "value": "value1"})
    assert variable.value == "value1"
    assert len(group.variables.list()) == 1

    variable.value = "new_value1"
    variable.save()
    variable = group.variables.get(variable.key)
    assert variable.value == "new_value1"

    variable.delete()
    assert len(group.variables.list()) == 0


def test_project_variables(project):
    variable = project.variables.create({"key": "key1", "value": "value1"})
    assert variable.value == "value1"
    assert len(project.variables.list()) == 1

    variable.value = "new_value1"
    variable.save()
    variable = project.variables.get(variable.key)
    assert variable.value == "new_value1"

    variable.delete()
    assert len(project.variables.list()) == 0



"""
GitLab API:
https://docs.gitlab.com/ee/api/instance_level_ci_variables.html
https://docs.gitlab.com/ee/api/project_level_variables.html
https://docs.gitlab.com/ee/api/group_level_variables.html
"""
from gitlab.base import *  # noqa
from gitlab.mixins import *  # noqa


class Variable(SaveMixin, ObjectDeleteMixin, RESTObject):
    _id_attr = "key"


class VariableManager(CRUDMixin, RESTManager):
    _path = "/admin/ci/variables"
    _obj_cls = Variable
    _create_attrs = (("key", "value"), ("protected", "variable_type", "masked"))
    _update_attrs = (("key", "value"), ("protected", "variable_type", "masked"))


class GroupVariable(SaveMixin, ObjectDeleteMixin, RESTObject):
    _id_attr = "key"


class GroupVariableManager(CRUDMixin, RESTManager):
    _path = "/groups/%(group_id)s/variables"
    _obj_cls = GroupVariable
    _from_parent_attrs = {"group_id": "id"}
    _create_attrs = (("key", "value"), ("protected", "variable_type", "masked"))
    _update_attrs = (("key", "value"), ("protected", "variable_type", "masked"))


class ProjectVariable(SaveMixin, ObjectDeleteMixin, RESTObject):
    _id_attr = "key"


class ProjectVariableManager(CRUDMixin, RESTManager):
    _path = "/projects/%(project_id)s/variables"
    _obj_cls = ProjectVariable
    _from_parent_attrs = {"project_id": "id"}
    _create_attrs = (
        ("key", "value"),
        ("protected", "variable_type", "masked", "environment_scope"),
    )
    _update_attrs = (
        ("key", "value"),
        ("protected", "variable_type", "masked", "environment_scope"),
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2013-2017 Gauvain Pocentek <gauvain@pocentek.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Wrapper for the GitLab API."""

import importlib
import time
import warnings

import requests
import requests.utils

import gitlab.config
from gitlab.const import *  # noqa
from gitlab.exceptions import *  # noqa
from gitlab import utils  # noqa

__title__ = "python-gitlab"
__version__ = "2.5.0"
__author__ = "Gauvain Pocentek"
__email__ = "gauvainpocentek@gmail.com"
__license__ = "LGPL3"
__copyright__ = "Copyright 2013-2019 Gauvain Pocentek"

warnings.filterwarnings("default", category=DeprecationWarning, module="^gitlab")

REDIRECT_MSG = (
    "python-gitlab detected an http to https redirection. You "
    "must update your GitLab URL to use https:// to avoid issues."
)

ALLOWED_KEYSET_ENDPOINTS = ["/projects"]


class Gitlab(object):
    """Represents a GitLab server connection.

    Args:
        url (str): The URL of the GitLab server.
        private_token (str): The user private token
        oauth_token (str): An oauth token
        job_token (str): A CI job token
        ssl_verify (bool|str): Whether SSL certificates should be validated. If
            the value is a string, it is the path to a CA file used for
            certificate validation.
        timeout (float): Timeout to use for requests to the GitLab server.
        http_username (str): Username for HTTP authentication
        http_password (str): Password for HTTP authentication
        api_version (str): Gitlab API version to use (support for 4 only)
        pagination (str): Can be set to 'keyset' to use keyset pagination
        order_by (str): Set order_by globally
    """

    def __init__(
        self,
        url,
        private_token=None,
        oauth_token=None,
        job_token=None,
        ssl_verify=True,
        http_username=None,
        http_password=None,
        timeout=None,
        api_version="4",
        session=None,
        per_page=None,
        pagination=None,
        order_by=None,
    ):

        self._api_version = str(api_version)
        self._server_version = self._server_revision = None
        self._base_url = url.rstrip("/")
        self._url = "%s/api/v%s" % (self._base_url, api_version)
        #: Timeout to use for requests to gitlab server
        self.timeout = timeout
        #: Headers that will be used in request to GitLab
        self.headers = {"User-Agent": "%s/%s" % (__title__, __version__)}

        #: Whether SSL certificates should be validated
        self.ssl_verify = ssl_verify

        self.private_token = private_token
        self.http_username = http_username
        self.http_password = http_password
        self.oauth_token = oauth_token
        self.job_token = job_token
        self._set_auth_info()

        #: Create a session object for requests
        self.session = session or requests.Session()

        self.per_page = per_page
        self.pagination = pagination
        self.order_by = order_by

        objects = importlib.import_module("gitlab.v%s.objects" % self._api_version)
        self._objects = objects

        self.broadcastmessages = objects.BroadcastMessageManager(self)
        self.deploykeys = objects.DeployKeyManager(self)
        self.deploytokens = objects.DeployTokenManager(self)
        self.geonodes = objects.GeoNodeManager(self)
        self.gitlabciymls = objects.GitlabciymlManager(self)
        self.gitignores = objects.GitignoreManager(self)
        self.groups = objects.GroupManager(self)
        self.hooks = objects.HookManager(self)
        self.issues = objects.IssueManager(self)
        self.ldapgroups = objects.LDAPGroupManager(self)
        self.licenses = objects.LicenseManager(self)
        self.namespaces = objects.NamespaceManager(self)
        self.mergerequests = objects.MergeRequestManager(self)
        self.notificationsettings = objects.NotificationSettingsManager(self)
        self.projects = objects.ProjectManager(self)
        self.runners = objects.RunnerManager(self)
        self.settings = objects.ApplicationSettingsManager(self)
        self.appearance = objects.ApplicationAppearanceManager(self)
        self.sidekiq = objects.SidekiqManager(self)
        self.snippets = objects.SnippetManager(self)
        self.users = objects.UserManager(self)
        self.todos = objects.TodoManager(self)
        self.dockerfiles = objects.DockerfileManager(self)
        self.events = objects.EventManager(self)
        self.audit_events = objects.AuditEventManager(self)
        self.features = objects.FeatureManager(self)
        self.pagesdomains = objects.PagesDomainManager(self)
        self.user_activities = objects.UserActivitiesManager(self)
        self.applications = objects.ApplicationManager(self)
        self.variables = objects.VariableManager(self)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.session.close()

    def __getstate__(self):
        state = self.__dict__.copy()
        state.pop("_objects")
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        objects = importlib.import_module("gitlab.v%s.objects" % self._api_version)
        self._objects = objects

    @property
    def url(self):
        """The user-provided server URL."""
        return self._base_url

    @property
    def api_url(self):
        """The computed API base URL."""
        return self._url

    @property
    def api_version(self):
        """The API version used (4 only)."""
        return self._api_version

    @classmethod
    def from_config(cls, gitlab_id=None, config_files=None):
        """Create a Gitlab connection from configuration files.

        Args:
            gitlab_id (str): ID of the configuration section.
            config_files list[str]: List of paths to configuration files.

        Returns:
            (gitlab.Gitlab): A Gitlab connection.

        Raises:
            gitlab.config.GitlabDataError: If the configuration is not correct.
        """
        config = gitlab.config.GitlabConfigParser(
            gitlab_id=gitlab_id, config_files=config_files
        )
        return cls(
            config.url,
            private_token=config.private_token,
            oauth_token=config.oauth_token,
            job_token=config.job_token,
            ssl_verify=config.ssl_verify,
            timeout=config.timeout,
            http_username=config.http_username,
            http_password=config.http_password,
            api_version=config.api_version,
            per_page=config.per_page,
            pagination=config.pagination,
            order_by=config.order_by,
        )

    def auth(self):
        """Performs an authentication using private token.

        The `user` attribute will hold a `gitlab.objects.CurrentUser` object on
        success.
        """
        self.user = self._objects.CurrentUserManager(self).get()

    def version(self):
        """Returns the version and revision of the gitlab server.

        Note that self.version and self.revision will be set on the gitlab
        object.

        Returns:
            tuple (str, str): The server version and server revision.
                              ('unknown', 'unknwown') if the server doesn't
                              perform as expected.
        """
        if self._server_version is None:
            try:
                data = self.http_get("/version")
                self._server_version = data["version"]
                self._server_revision = data["revision"]
            except Exception:
                self._server_version = self._server_revision = "unknown"

        return self._server_version, self._server_revision

    @on_http_error(GitlabVerifyError)
    def lint(self, content, **kwargs):
        """Validate a gitlab CI configuration.

        Args:
            content (txt): The .gitlab-ci.yml content
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabVerifyError: If the validation could not be done

        Returns:
            tuple: (True, []) if the file is valid, (False, errors(list))
                otherwise
        """
        post_data = {"content": content}
        data = self.http_post("/ci/lint", post_data=post_data, **kwargs)
        return (data["status"] == "valid", data["errors"])

    @on_http_error(GitlabMarkdownError)
    def markdown(self, text, gfm=False, project=None, **kwargs):
        """Render an arbitrary Markdown document.

        Args:
            text (str): The markdown text to render
            gfm (bool): Render text using GitLab Flavored Markdown. Default is
                False
            project (str): Full path of a project used a context when `gfm` is
                True
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabMarkdownError: If the server cannot perform the request

        Returns:
            str: The HTML rendering of the markdown text.
        """
        post_data = {"text": text, "gfm": gfm}
        if project is not None:
            post_data["project"] = project
        data = self.http_post("/markdown", post_data=post_data, **kwargs)
        return data["html"]

    @on_http_error(GitlabLicenseError)
    def get_license(self, **kwargs):
        """Retrieve information about the current license.

        Args:
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabGetError: If the server cannot perform the request

        Returns:
            dict: The current license information
        """
        return self.http_get("/license", **kwargs)

    @on_http_error(GitlabLicenseError)
    def set_license(self, license, **kwargs):
        """Add a new license.

        Args:
            license (str): The license string
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabPostError: If the server cannot perform the request

        Returns:
            dict: The new license information
        """
        data = {"license": license}
        return self.http_post("/license", post_data=data, **kwargs)

    def _construct_url(self, id_, obj, parameters, action=None):
        if "next_url" in parameters:
            return parameters["next_url"]
        args = utils.sanitize_parameters(parameters)

        url_attr = "_url"
        if action is not None:
            attr = "_%s_url" % action
            if hasattr(obj, attr):
                url_attr = attr
        obj_url = getattr(obj, url_attr)
        url = obj_url % args

        if id_ is not None:
            return "%s/%s" % (url, str(id_))
        else:
            return url

    def _set_auth_info(self):
        tokens = [
            token
            for token in [self.private_token, self.oauth_token, self.job_token]
            if token
        ]
        if len(tokens) > 1:
            raise ValueError(
                "Only one of private_token, oauth_token or job_token should "
                "be defined"
            )
        if (self.http_username and not self.http_password) or (
            not self.http_username and self.http_password
        ):
            raise ValueError(
                "Both http_username and http_password should " "be defined"
            )
        if self.oauth_token and self.http_username:
            raise ValueError(
                "Only one of oauth authentication or http "
                "authentication should be defined"
            )

        self._http_auth = None
        if self.private_token:
            self.headers.pop("Authorization", None)
            self.headers["PRIVATE-TOKEN"] = self.private_token
            self.headers.pop("JOB-TOKEN", None)

        if self.oauth_token:
            self.headers["Authorization"] = "Bearer %s" % self.oauth_token
            self.headers.pop("PRIVATE-TOKEN", None)
            self.headers.pop("JOB-TOKEN", None)

        if self.job_token:
            self.headers.pop("Authorization", None)
            self.headers.pop("PRIVATE-TOKEN", None)
            self.headers["JOB-TOKEN"] = self.job_token

        if self.http_username:
            self._http_auth = requests.auth.HTTPBasicAuth(
                self.http_username, self.http_password
            )

    def enable_debug(self):
        import logging

        try:
            from http.client import HTTPConnection  # noqa
        except ImportError:
            from httplib import HTTPConnection  # noqa

        HTTPConnection.debuglevel = 1
        logging.basicConfig()
        logging.getLogger().setLevel(logging.DEBUG)
        requests_log = logging.getLogger("requests.packages.urllib3")
        requests_log.setLevel(logging.DEBUG)
        requests_log.propagate = True

    def _create_headers(self, content_type=None):
        request_headers = self.headers.copy()
        if content_type is not None:
            request_headers["Content-type"] = content_type
        return request_headers

    def _get_session_opts(self, content_type):
        return {
            "headers": self._create_headers(content_type),
            "auth": self._http_auth,
            "timeout": self.timeout,
            "verify": self.ssl_verify,
        }

    def _build_url(self, path):
        """Returns the full url from path.

        If path is already a url, return it unchanged. If it's a path, append
        it to the stored url.

        Returns:
            str: The full URL
        """
        if path.startswith("http://") or path.startswith("https://"):
            return path
        else:
            return "%s%s" % (self._url, path)

    def _check_redirects(self, result):
        # Check the requests history to detect http to https redirections.
        # If the initial verb is POST, the next request will use a GET request,
        # leading to an unwanted behaviour.
        # If the initial verb is PUT, the data will not be send with the next
        # request.
        # If we detect a redirection to https with a POST or a PUT request, we
        # raise an exception with a useful error message.
        if result.history and self._base_url.startswith("http:"):
            for item in result.history:
                if item.status_code not in (301, 302):
                    continue
                # GET methods can be redirected without issue
                if item.request.method == "GET":
                    continue
                # Did we end-up with an https:// URL?
                location = item.headers.get("Location", None)
                if location and location.startswith("https://"):
                    raise RedirectError(REDIRECT_MSG)

    def http_request(
        self,
        verb,
        path,
        query_data=None,
        post_data=None,
        streamed=False,
        files=None,
        **kwargs
    ):
        """Make an HTTP request to the Gitlab server.

        Args:
            verb (str): The HTTP method to call ('get', 'post', 'put',
                        'delete')
            path (str): Path or full URL to query ('/projects' or
                        'http://whatever/v4/api/projecs')
            query_data (dict): Data to send as query parameters
            post_data (dict): Data to send in the body (will be converted to
                              json)
            streamed (bool): Whether the data should be streamed
            files (dict): The files to send to the server
            **kwargs: Extra options to send to the server (e.g. sudo)

        Returns:
            A requests result object.

        Raises:
            GitlabHttpError: When the return code is not 2xx
        """
        query_data = query_data or {}
        url = self._build_url(path)

        params = {}
        utils.copy_dict(params, query_data)

        # Deal with kwargs: by default a user uses kwargs to send data to the
        # gitlab server, but this generates problems (python keyword conflicts
        # and python-gitlab/gitlab conflicts).
        # So we provide a `query_parameters` key: if it's there we use its dict
        # value as arguments for the gitlab server, and ignore the other
        # arguments, except pagination ones (per_page and page)
        if "query_parameters" in kwargs:
            utils.copy_dict(params, kwargs["query_parameters"])
            for arg in ("per_page", "page"):
                if arg in kwargs:
                    params[arg] = kwargs[arg]
        else:
            utils.copy_dict(params, kwargs)

        opts = self._get_session_opts(content_type="application/json")

        verify = opts.pop("verify")
        timeout = opts.pop("timeout")
        # If timeout was passed into kwargs, allow it to override the default
        timeout = kwargs.get("timeout", timeout)

        # We need to deal with json vs. data when uploading files
        if files:
            data = post_data
            json = None
            del opts["headers"]["Content-type"]
        else:
            json = post_data
            data = None

        # Requests assumes that `.` should not be encoded as %2E and will make
        # changes to urls using this encoding. Using a prepped request we can
        # get the desired behavior.
        # The Requests behavior is right but it seems that web servers don't
        # always agree with this decision (this is the case with a default
        # gitlab installation)
        req = requests.Request(
            verb, url, json=json, data=data, params=params, files=files, **opts
        )
        prepped = self.session.prepare_request(req)
        prepped.url = utils.sanitized_url(prepped.url)
        settings = self.session.merge_environment_settings(
            prepped.url, {}, streamed, verify, None
        )

        # obey the rate limit by default
        obey_rate_limit = kwargs.get("obey_rate_limit", True)
        # do not retry transient errors by default
        retry_transient_errors = kwargs.get("retry_transient_errors", False)

        # set max_retries to 10 by default, disable by setting it to -1
        max_retries = kwargs.get("max_retries", 10)
        cur_retries = 0

        while True:
            result = self.session.send(prepped, timeout=timeout, **settings)

            self._check_redirects(result)

            if 200 <= result.status_code < 300:
                return result

            if (429 == result.status_code and obey_rate_limit) or (
                result.status_code in [500, 502, 503, 504] and retry_transient_errors
            ):
                if max_retries == -1 or cur_retries < max_retries:
                    wait_time = 2 ** cur_retries * 0.1
                    if "Retry-After" in result.headers:
                        wait_time = int(result.headers["Retry-After"])
                    cur_retries += 1
                    time.sleep(wait_time)
                    continue

            error_message = result.content
            try:
                error_json = result.json()
                for k in ("message", "error"):
                    if k in error_json:
                        error_message = error_json[k]
            except (KeyError, ValueError, TypeError):
                pass

            if result.status_code == 401:
                raise GitlabAuthenticationError(
                    response_code=result.status_code,
                    error_message=error_message,
                    response_body=result.content,
                )

            raise GitlabHttpError(
                response_code=result.status_code,
                error_message=error_message,
                response_body=result.content,
            )

    def http_get(self, path, query_data=None, streamed=False, raw=False, **kwargs):
        """Make a GET request to the Gitlab server.

        Args:
            path (str): Path or full URL to query ('/projects' or
                        'http://whatever/v4/api/projecs')
            query_data (dict): Data to send as query parameters
            streamed (bool): Whether the data should be streamed
            raw (bool): If True do not try to parse the output as json
            **kwargs: Extra options to send to the server (e.g. sudo)

        Returns:
            A requests result object is streamed is True or the content type is
            not json.
            The parsed json data otherwise.

        Raises:
            GitlabHttpError: When the return code is not 2xx
            GitlabParsingError: If the json data could not be parsed
        """
        query_data = query_data or {}
        result = self.http_request(
            "get", path, query_data=query_data, streamed=streamed, **kwargs
        )

        if (
            result.headers["Content-Type"] == "application/json"
            and not streamed
            and not raw
        ):
            try:
                return result.json()
            except Exception as e:
                raise GitlabParsingError(
                    error_message="Failed to parse the server message"
                ) from e
        else:
            return result

    def http_list(self, path, query_data=None, as_list=None, **kwargs):
        """Make a GET request to the Gitlab server for list-oriented queries.

        Args:
            path (str): Path or full URL to query ('/projects' or
                        'http://whatever/v4/api/projects')
            query_data (dict): Data to send as query parameters
            **kwargs: Extra options to send to the server (e.g. sudo, page,
                      per_page)

        Returns:
            list: A list of the objects returned by the server. If `as_list` is
            False and no pagination-related arguments (`page`, `per_page`,
            `all`) are defined then a GitlabList object (generator) is returned
            instead. This object will make API calls when needed to fetch the
            next items from the server.

        Raises:
            GitlabHttpError: When the return code is not 2xx
            GitlabParsingError: If the json data could not be parsed
        """
        query_data = query_data or {}

        # In case we want to change the default behavior at some point
        as_list = True if as_list is None else as_list

        get_all = kwargs.pop("all", False)
        url = self._build_url(path)

        page = kwargs.get("page")

        if get_all is True and as_list is True:
            return list(GitlabList(self, url, query_data, **kwargs))

        if page or as_list is True:
            # pagination requested, we return a list
            return list(GitlabList(self, url, query_data, get_next=False, **kwargs))

        # No pagination, generator requested
        return GitlabList(self, url, query_data, **kwargs)

    def http_post(self, path, query_data=None, post_data=None, files=None, **kwargs):
        """Make a POST request to the Gitlab server.

        Args:
            path (str): Path or full URL to query ('/projects' or
                        'http://whatever/v4/api/projecs')
            query_data (dict): Data to send as query parameters
            post_data (dict): Data to send in the body (will be converted to
                              json)
            files (dict): The files to send to the server
            **kwargs: Extra options to send to the server (e.g. sudo)

        Returns:
            The parsed json returned by the server if json is return, else the
            raw content

        Raises:
            GitlabHttpError: When the return code is not 2xx
            GitlabParsingError: If the json data could not be parsed
        """
        query_data = query_data or {}
        post_data = post_data or {}

        result = self.http_request(
            "post",
            path,
            query_data=query_data,
            post_data=post_data,
            files=files,
            **kwargs
        )
        try:
            if result.headers.get("Content-Type", None) == "application/json":
                return result.json()
        except Exception as e:
            raise GitlabParsingError(
                error_message="Failed to parse the server message"
            ) from e
        return result

    def http_put(self, path, query_data=None, post_data=None, files=None, **kwargs):
        """Make a PUT request to the Gitlab server.

        Args:
            path (str): Path or full URL to query ('/projects' or
                        'http://whatever/v4/api/projecs')
            query_data (dict): Data to send as query parameters
            post_data (dict): Data to send in the body (will be converted to
                              json)
            files (dict): The files to send to the server
            **kwargs: Extra options to send to the server (e.g. sudo)

        Returns:
            The parsed json returned by the server.

        Raises:
            GitlabHttpError: When the return code is not 2xx
            GitlabParsingError: If the json data could not be parsed
        """
        query_data = query_data or {}
        post_data = post_data or {}

        result = self.http_request(
            "put",
            path,
            query_data=query_data,
            post_data=post_data,
            files=files,
            **kwargs
        )
        try:
            return result.json()
        except Exception as e:
            raise GitlabParsingError(
                error_message="Failed to parse the server message"
            ) from e

    def http_delete(self, path, **kwargs):
        """Make a PUT request to the Gitlab server.

        Args:
            path (str): Path or full URL to query ('/projects' or
                        'http://whatever/v4/api/projecs')
            **kwargs: Extra options to send to the server (e.g. sudo)

        Returns:
            The requests object.

        Raises:
            GitlabHttpError: When the return code is not 2xx
        """
        return self.http_request("delete", path, **kwargs)

    @on_http_error(GitlabSearchError)
    def search(self, scope, search, **kwargs):
        """Search GitLab resources matching the provided string.'

        Args:
            scope (str): Scope of the search
            search (str): Search string
            **kwargs: Extra options to send to the server (e.g. sudo)

        Raises:
            GitlabAuthenticationError: If authentication is not correct
            GitlabSearchError: If the server failed to perform the request

        Returns:
            GitlabList: A list of dicts describing the resources found.
        """
        data = {"scope": scope, "search": search}
        return self.http_list("/search", query_data=data, **kwargs)


class GitlabList(object):
    """Generator representing a list of remote objects.

    The object handles the links returned by a query to the API, and will call
    the API again when needed.
    """

    def __init__(self, gl, url, query_data, get_next=True, **kwargs):
        self._gl = gl

        # Preserve kwargs for subsequent queries
        self._kwargs = kwargs.copy()

        self._query(url, query_data, **self._kwargs)
        self._get_next = get_next

    def _query(self, url, query_data=None, **kwargs):
        query_data = query_data or {}
        result = self._gl.http_request("get", url, query_data=query_data, **kwargs)
        try:
            links = result.links
            if links:
                next_url = links["next"]["url"]
            else:
                next_url = requests.utils.parse_header_links(result.headers["links"])[
                    0
                ]["url"]
            self._next_url = next_url
        except KeyError:
            self._next_url = None
        self._current_page = result.headers.get("X-Page")
        self._prev_page = result.headers.get("X-Prev-Page")
        self._next_page = result.headers.get("X-Next-Page")
        self._per_page = result.headers.get("X-Per-Page")
        self._total_pages = result.headers.get("X-Total-Pages")
        self._total = result.headers.get("X-Total")

        try:
            self._data = result.json()
        except Exception as e:
            raise GitlabParsingError(
                error_message="Failed to parse the server message"
            ) from e

        self._current = 0

    @property
    def current_page(self):
        """The current page number."""
        return int(self._current_page)

    @property
    def prev_page(self):
        """The previous page number.

        If None, the current page is the first.
        """
        return int(self._prev_page) if self._prev_page else None

    @property
    def next_page(self):
        """The next page number.

        If None, the current page is the last.
        """
        return int(self._next_page) if self._next_page else None

    @property
    def per_page(self):
        """The number of items per page."""
        return int(self._per_page)

    @property
    def total_pages(self):
        """The total number of pages."""
        return int(self._total_pages)

    @property
    def total(self):
        """The total number of items."""
        return int(self._total)

    def __iter__(self):
        return self

    def __len__(self):
        return int(self._total)

    def __next__(self):
        return self.next()

    def next(self):
        try:
            item = self._data[self._current]
            self._current += 1
            return item
        except IndexError:
            pass

        if self._next_url and self._get_next is True:
            self._query(self._next_url, **self._kwargs)
            return self.next()

        raise StopIteration



"""
GitLab API: https://docs.gitlab.com/ce/api/projects.html
"""

import pytest
import responses

from gitlab.v4.objects import Project


project_content = {"name": "name", "id": 1}


@pytest.fixture
def resp_get_project():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1",
            json=project_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_list_projects():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects",
            json=[project_content],
            content_type="application/json",
            status=200,
        )
        yield rsps


def test_get_project(gl, resp_get_project):
    data = gl.projects.get(1)
    assert isinstance(data, Project)
    assert data.name == "name"
    assert data.id == 1


def test_list_projects(gl, resp_list_projects):
    projects = gl.projects.list()
    assert isinstance(projects[0], Project)
    assert projects[0].name == "name"


@pytest.mark.skip(reason="missing test")
def test_list_user_projects(gl):
    pass


@pytest.mark.skip(reason="missing test")
def test_list_user_starred_projects(gl):
    pass


@pytest.mark.skip(reason="missing test")
def test_list_project_users(gl):
    pass


@pytest.mark.skip(reason="missing test")
def test_create_project(gl):
    pass


@pytest.mark.skip(reason="missing test")
def test_create_user_project(gl):
    pass


@pytest.mark.skip(reason="missing test")
def test_update_project(gl):
    pass


@pytest.mark.skip(reason="missing test")
def test_fork_project(gl):
    pass


@pytest.mark.skip(reason="missing test")
def test_list_project_forks(gl):
    pass


@pytest.mark.skip(reason="missing test")
def test_star_project(gl):
    pass


@pytest.mark.skip(reason="missing test")
def test_unstar_project(gl):
    pass


@pytest.mark.skip(reason="missing test")
def test_list_project_starrers(gl):
    pass


@pytest.mark.skip(reason="missing test")
def test_get_project_languages(gl):
    pass


@pytest.mark.skip(reason="missing test")
def test_archive_project(gl):
    pass


@pytest.mark.skip(reason="missing test")
def test_unarchive_project(gl):
    pass


@pytest.mark.skip(reason="missing test")
def test_remove_project(gl):
    pass


@pytest.mark.skip(reason="missing test")
def test_restore_project(gl):
    pass


@pytest.mark.skip(reason="missing test")
def test_upload_file(gl):
    pass


@pytest.mark.skip(reason="missing test")
def test_share_project(gl):
    pass


@pytest.mark.skip(reason="missing test")
def test_delete_shared_project_link(gl):
    pass


@pytest.mark.skip(reason="missing test")
def test_list_project_hooks(gl):
    pass


@pytest.mark.skip(reason="missing test")
def test_get_project_hook(gl):
    pass


@pytest.mark.skip(reason="missing test")
def test_create_project_hook(gl):
    pass


@pytest.mark.skip(reason="missing test")
def test_update_project_hook(gl):
    pass


@pytest.mark.skip(reason="missing test")
def test_delete_project_hook(gl):
    pass


@pytest.mark.skip(reason="missing test")
def test_create_forked_from_relationship(gl):
    pass


@pytest.mark.skip(reason="missing test")
def test_delete_forked_from_relationship(gl):
    pass


@pytest.mark.skip(reason="missing test")
def test_search_projects_by_name(gl):
    pass


@pytest.mark.skip(reason="missing test")
def test_project_housekeeping(gl):
    pass


@pytest.mark.skip(reason="missing test")
def test_get_project_push_rules(gl):
    pass


@pytest.mark.skip(reason="missing test")
def test_create_project_push_rule(gl):
    pass


@pytest.mark.skip(reason="missing test")
def test_update_project_push_rule(gl):
    pass


@pytest.mark.skip(reason="missing test")
def test_delete_project_push_rule(gl):
    pass


@pytest.mark.skip(reason="missing test")
def test_transfer_project(gl):
    pass


@pytest.mark.skip(reason="missing test")
def test_project_pull_mirror(gl):
    pass


@pytest.mark.skip(reason="missing test")
def test_project_snapshot(gl):
    pass



# -*- coding: utf-8 -*-
#
# Copyright (C) 2013-2017 Gauvain Pocentek <gauvain@pocentek.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import importlib


class RESTObject(object):
    """Represents an object built from server data.

    It holds the attributes know from the server, and the updated attributes in
    another. This allows smart updates, if the object allows it.

    You can redefine ``_id_attr`` in child classes to specify which attribute
    must be used as uniq ID. ``None`` means that the object can be updated
    without ID in the url.
    """

    _id_attr = "id"

    def __init__(self, manager, attrs):
        self.__dict__.update(
            {
                "manager": manager,
                "_attrs": attrs,
                "_updated_attrs": {},
                "_module": importlib.import_module(self.__module__),
            }
        )
        self.__dict__["_parent_attrs"] = self.manager.parent_attrs
        self._create_managers()

    def __getstate__(self):
        state = self.__dict__.copy()
        module = state.pop("_module")
        state["_module_name"] = module.__name__
        return state

    def __setstate__(self, state):
        module_name = state.pop("_module_name")
        self.__dict__.update(state)
        self.__dict__["_module"] = importlib.import_module(module_name)

    def __getattr__(self, name):
        try:
            return self.__dict__["_updated_attrs"][name]
        except KeyError:
            try:
                value = self.__dict__["_attrs"][name]

                # If the value is a list, we copy it in the _updated_attrs dict
                # because we are not able to detect changes made on the object
                # (append, insert, pop, ...). Without forcing the attr
                # creation __setattr__ is never called, the list never ends up
                # in the _updated_attrs dict, and the update() and save()
                # method never push the new data to the server.
                # See https://github.com/python-gitlab/python-gitlab/issues/306
                #
                # note: _parent_attrs will only store simple values (int) so we
                # don't make this check in the next except block.
                if isinstance(value, list):
                    self.__dict__["_updated_attrs"][name] = value[:]
                    return self.__dict__["_updated_attrs"][name]

                return value

            except KeyError:
                try:
                    return self.__dict__["_parent_attrs"][name]
                except KeyError:
                    raise AttributeError(name)

    def __setattr__(self, name, value):
        self.__dict__["_updated_attrs"][name] = value

    def __str__(self):
        data = self._attrs.copy()
        data.update(self._updated_attrs)
        return "%s => %s" % (type(self), data)

    def __repr__(self):
        if self._id_attr:
            return "<%s %s:%s>" % (
                self.__class__.__name__,
                self._id_attr,
                self.get_id(),
            )
        else:
            return "<%s>" % self.__class__.__name__

    def __eq__(self, other):
        if self.get_id() and other.get_id():
            return self.get_id() == other.get_id()
        return super(RESTObject, self) == other

    def __ne__(self, other):
        if self.get_id() and other.get_id():
            return self.get_id() != other.get_id()
        return super(RESTObject, self) != other

    def __dir__(self):
        return super(RESTObject, self).__dir__() + list(self.attributes)

    def __hash__(self):
        if not self.get_id():
            return super(RESTObject, self).__hash__()
        return hash(self.get_id())

    def _create_managers(self):
        managers = getattr(self, "_managers", None)
        if managers is None:
            return

        for attr, cls_name in self._managers:
            cls = getattr(self._module, cls_name)
            manager = cls(self.manager.gitlab, parent=self)
            self.__dict__[attr] = manager

    def _update_attrs(self, new_attrs):
        self.__dict__["_updated_attrs"] = {}
        self.__dict__["_attrs"].update(new_attrs)

    def get_id(self):
        """Returns the id of the resource."""
        if self._id_attr is None or not hasattr(self, self._id_attr):
            return None
        return getattr(self, self._id_attr)

    @property
    def attributes(self):
        d = self.__dict__["_updated_attrs"].copy()
        d.update(self.__dict__["_attrs"])
        d.update(self.__dict__["_parent_attrs"])
        return d


class RESTObjectList(object):
    """Generator object representing a list of RESTObject's.

    This generator uses the Gitlab pagination system to fetch new data when
    required.

    Note: you should not instanciate such objects, they are returned by calls
    to RESTManager.list()

    Args:
        manager: Manager to attach to the created objects
        obj_cls: Type of objects to create from the json data
        _list: A GitlabList object
    """

    def __init__(self, manager, obj_cls, _list):
        """Creates an objects list from a GitlabList.

        You should not create objects of this type, but use managers list()
        methods instead.

        Args:
            manager: the RESTManager to attach to the objects
            obj_cls: the class of the created objects
            _list: the GitlabList holding the data
        """
        self.manager = manager
        self._obj_cls = obj_cls
        self._list = _list

    def __iter__(self):
        return self

    def __len__(self):
        return len(self._list)

    def __next__(self):
        return self.next()

    def next(self):
        data = self._list.next()
        return self._obj_cls(self.manager, data)

    @property
    def current_page(self):
        """The current page number."""
        return self._list.current_page

    @property
    def prev_page(self):
        """The previous page number.

        If None, the current page is the first.
        """
        return self._list.prev_page

    @property
    def next_page(self):
        """The next page number.

        If None, the current page is the last.
        """
        return self._list.next_page

    @property
    def per_page(self):
        """The number of items per page."""
        return self._list.per_page

    @property
    def total_pages(self):
        """The total number of pages."""
        return self._list.total_pages

    @property
    def total(self):
        """The total number of items."""
        return self._list.total


class RESTManager(object):
    """Base class for CRUD operations on objects.

    Derived class must define ``_path`` and ``_obj_cls``.

    ``_path``: Base URL path on which requests will be sent (e.g. '/projects')
    ``_obj_cls``: The class of objects that will be created
    """

    _path = None
    _obj_cls = None

    def __init__(self, gl, parent=None):
        """REST manager constructor.

        Args:
            gl (Gitlab): :class:`~gitlab.Gitlab` connection to use to make
                         requests.
            parent: REST object to which the manager is attached.
        """
        self.gitlab = gl
        self._parent = parent  # for nested managers
        self._computed_path = self._compute_path()

    @property
    def parent_attrs(self):
        return self._parent_attrs

    def _compute_path(self, path=None):
        self._parent_attrs = {}
        if path is None:
            path = self._path
        if self._parent is None or not hasattr(self, "_from_parent_attrs"):
            return path

        data = {
            self_attr: getattr(self._parent, parent_attr, None)
            for self_attr, parent_attr in self._from_parent_attrs.items()
        }
        self._parent_attrs = data
        return path % data

    @property
    def path(self):
        return self._computed_path



# -*- coding: utf-8 -*-
#
# Copyright (C) 2016-2017 Gauvain Pocentek <gauvain@pocentek.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import unittest

import mock
import io

from gitlab import config
import pytest


valid_config = u"""[global]
default = one
ssl_verify = true
timeout = 2

[one]
url = http://one.url
private_token = ABCDEF

[two]
url = https://two.url
private_token = GHIJKL
ssl_verify = false
timeout = 10

[three]
url = https://three.url
private_token = MNOPQR
ssl_verify = /path/to/CA/bundle.crt
per_page = 50

[four]
url = https://four.url
oauth_token = STUV
"""

no_default_config = u"""[global]
[there]
url = http://there.url
private_token = ABCDEF
"""

missing_attr_config = u"""[global]
[one]
url = http://one.url

[two]
private_token = ABCDEF

[three]
meh = hem

[four]
url = http://four.url
private_token = ABCDEF
per_page = 200
"""


@mock.patch.dict(os.environ, {"PYTHON_GITLAB_CFG": "/some/path"})
def test_env_config_present():
    assert ["/some/path"] == config._env_config()


@mock.patch.dict(os.environ, {}, clear=True)
def test_env_config_missing():
    assert [] == config._env_config()


@mock.patch("os.path.exists")
def test_missing_config(path_exists):
    path_exists.return_value = False
    with pytest.raises(config.GitlabConfigMissingError):
        config.GitlabConfigParser("test")


@mock.patch("os.path.exists")
@mock.patch("builtins.open")
def test_invalid_id(m_open, path_exists):
    fd = io.StringIO(no_default_config)
    fd.close = mock.Mock(return_value=None)
    m_open.return_value = fd
    path_exists.return_value = True
    config.GitlabConfigParser("there")
    with pytest.raises(config.GitlabIDError):
        config.GitlabConfigParser()

    fd = io.StringIO(valid_config)
    fd.close = mock.Mock(return_value=None)
    m_open.return_value = fd
    with pytest.raises(config.GitlabDataError):
        config.GitlabConfigParser(gitlab_id="not_there")


@mock.patch("os.path.exists")
@mock.patch("builtins.open")
def test_invalid_data(m_open, path_exists):
    fd = io.StringIO(missing_attr_config)
    fd.close = mock.Mock(return_value=None, side_effect=lambda: fd.seek(0))
    m_open.return_value = fd
    path_exists.return_value = True

    config.GitlabConfigParser("one")
    config.GitlabConfigParser("one")
    with pytest.raises(config.GitlabDataError):
        config.GitlabConfigParser(gitlab_id="two")
    with pytest.raises(config.GitlabDataError):
        config.GitlabConfigParser(gitlab_id="three")
    with pytest.raises(config.GitlabDataError) as emgr:
        config.GitlabConfigParser("four")
    assert "Unsupported per_page number: 200" == emgr.value.args[0]


@mock.patch("os.path.exists")
@mock.patch("builtins.open")
def test_valid_data(m_open, path_exists):
    fd = io.StringIO(valid_config)
    fd.close = mock.Mock(return_value=None)
    m_open.return_value = fd
    path_exists.return_value = True

    cp = config.GitlabConfigParser()
    assert "one" == cp.gitlab_id
    assert "http://one.url" == cp.url
    assert "ABCDEF" == cp.private_token
    assert None == cp.oauth_token
    assert 2 == cp.timeout
    assert True == cp.ssl_verify
    assert cp.per_page is None

    fd = io.StringIO(valid_config)
    fd.close = mock.Mock(return_value=None)
    m_open.return_value = fd
    cp = config.GitlabConfigParser(gitlab_id="two")
    assert "two" == cp.gitlab_id
    assert "https://two.url" == cp.url
    assert "GHIJKL" == cp.private_token
    assert None == cp.oauth_token
    assert 10 == cp.timeout
    assert False == cp.ssl_verify

    fd = io.StringIO(valid_config)
    fd.close = mock.Mock(return_value=None)
    m_open.return_value = fd
    cp = config.GitlabConfigParser(gitlab_id="three")
    assert "three" == cp.gitlab_id
    assert "https://three.url" == cp.url
    assert "MNOPQR" == cp.private_token
    assert None == cp.oauth_token
    assert 2 == cp.timeout
    assert "/path/to/CA/bundle.crt" == cp.ssl_verify
    assert 50 == cp.per_page

    fd = io.StringIO(valid_config)
    fd.close = mock.Mock(return_value=None)
    m_open.return_value = fd
    cp = config.GitlabConfigParser(gitlab_id="four")
    assert "four" == cp.gitlab_id
    assert "https://four.url" == cp.url
    assert None == cp.private_token
    assert "STUV" == cp.oauth_token
    assert 2 == cp.timeout
    assert True == cp.ssl_verify



"""
GitLab API: https://docs.gitlab.com/ce/api/groups.html
"""

import pytest
import responses

import gitlab


@pytest.fixture
def resp_groups():
    content = {"name": "name", "id": 1, "path": "path"}

    with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/groups/1",
            json=content,
            content_type="application/json",
            status=200,
        )
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/groups",
            json=[content],
            content_type="application/json",
            status=200,
        )
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/groups",
            json=content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_create_import(accepted_content):
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/groups/import",
            json=accepted_content,
            content_type="application/json",
            status=202,
        )
        yield rsps


def test_get_group(gl, resp_groups):
    data = gl.groups.get(1)
    assert isinstance(data, gitlab.v4.objects.Group)
    assert data.name == "name"
    assert data.path == "path"
    assert data.id == 1


def test_create_group(gl, resp_groups):
    name, path = "name", "path"
    data = gl.groups.create({"name": name, "path": path})
    assert isinstance(data, gitlab.v4.objects.Group)
    assert data.name == name
    assert data.path == path


def test_create_group_export(group, resp_export):
    export = group.exports.create()
    assert export.message == "202 Accepted"


@pytest.mark.skip("GitLab API endpoint not implemented")
def test_refresh_group_export_status(group, resp_export):
    export = group.exports.create()
    export.refresh()
    assert export.export_status == "finished"


def test_download_group_export(group, resp_export, binary_content):
    export = group.exports.create()
    download = export.download()
    assert isinstance(download, bytes)
    assert download == binary_content


def test_import_group(gl, resp_create_import):
    group_import = gl.groups.import_group("file", "api-group", "API Group")
    assert group_import["message"] == "202 Accepted"


@pytest.mark.skip("GitLab API endpoint not implemented")
def test_refresh_group_import_status(group, resp_groups):
    group_import = group.imports.get()
    group_import.refresh()
    assert group_import.import_status == "finished"



"""
GitLab API: https://docs.gitlab.com/ce/api/project_statistics.html
"""
import pytest
import responses

from gitlab.v4.objects import ProjectAdditionalStatistics


@pytest.fixture
def resp_project_statistics():
    content = {"fetches": {"total": 50, "days": [{"count": 10, "date": "2018-01-10"}]}}

    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/statistics",
            json=content,
            content_type="application/json",
            status=200,
        )
        yield rsps


def test_project_additional_statistics(project, resp_project_statistics):
    statistics = project.additionalstatistics.get()
    assert isinstance(statistics, ProjectAdditionalStatistics)
    assert statistics.fetches["total"] == 50



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 Gauvain Pocentek <gauvain@pocentek.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


class GitlabAttribute(object):
    def __init__(self, value=None):
        self._value = value

    def get(self):
        return self._value

    def set_from_cli(self, cli_value):
        self._value = cli_value

    def get_for_api(self):
        return self._value


class ListAttribute(GitlabAttribute):
    def set_from_cli(self, cli_value):
        if not cli_value.strip():
            self._value = []
        else:
            self._value = [item.strip() for item in cli_value.split(",")]

    def get_for_api(self):
        # Do not comma-split single value passed as string
        if isinstance(self._value, str):
            return self._value

        return ",".join(self._value)


class LowercaseStringAttribute(GitlabAttribute):
    def get_for_api(self):
        return str(self._value).lower()


class FileAttribute(GitlabAttribute):
    def get_file_name(self, attr_name=None):
        return attr_name


class ImageAttribute(FileAttribute):
    def get_file_name(self, attr_name=None):
        return "%s.png" % attr_name if attr_name else "image.png"



# -*- coding: utf-8 -*-
#
# Copyright (C) 2016-2017 Gauvain Pocentek <gauvain@pocentek.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from urllib.parse import urlparse


class _StdoutStream(object):
    def __call__(self, chunk):
        print(chunk)


def response_content(response, streamed, action, chunk_size):
    if streamed is False:
        return response.content

    if action is None:
        action = _StdoutStream()

    for chunk in response.iter_content(chunk_size=chunk_size):
        if chunk:
            action(chunk)


def copy_dict(dest, src):
    for k, v in src.items():
        if isinstance(v, dict):
            # Transform dict values to new attributes. For example:
            # custom_attributes: {'foo', 'bar'} =>
            #   "custom_attributes['foo']": "bar"
            for dict_k, dict_v in v.items():
                dest["%s[%s]" % (k, dict_k)] = dict_v
        else:
            dest[k] = v


def clean_str_id(id):
    return id.replace("/", "%2F").replace("#", "%23")


def sanitize_parameters(value):
    if isinstance(value, dict):
        return dict((k, sanitize_parameters(v)) for k, v in value.items())
    if isinstance(value, str):
        return value.replace("/", "%2F")
    return value


def sanitized_url(url):
    parsed = urlparse(url)
    new_path = parsed.path.replace(".", "%2E")
    return parsed._replace(path=new_path).geturl()


def remove_none_from_dict(data):
    return {k: v for k, v in data.items() if v is not None}



"""
GitLab API: https://docs.gitlab.com/ce/api/packages.html
"""


def test_list_project_packages(project):
    packages = project.packages.list()
    assert isinstance(packages, list)


def test_list_group_packages(group):
    packages = group.packages.list()
    assert isinstance(packages, list)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 Gauvain Pocentek <gauvain@pocentek.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gitlab import types


def test_gitlab_attribute_get():
    o = types.GitlabAttribute("whatever")
    assert o.get() == "whatever"

    o.set_from_cli("whatever2")
    assert o.get() == "whatever2"
    assert o.get_for_api() == "whatever2"

    o = types.GitlabAttribute()
    assert o._value is None


def test_list_attribute_input():
    o = types.ListAttribute()
    o.set_from_cli("foo,bar,baz")
    assert o.get() == ["foo", "bar", "baz"]

    o.set_from_cli("foo")
    assert o.get() == ["foo"]


def test_list_attribute_empty_input():
    o = types.ListAttribute()
    o.set_from_cli("")
    assert o.get() == []

    o.set_from_cli("  ")
    assert o.get() == []


def test_list_attribute_get_for_api_from_cli():
    o = types.ListAttribute()
    o.set_from_cli("foo,bar,baz")
    assert o.get_for_api() == "foo,bar,baz"


def test_list_attribute_get_for_api_from_list():
    o = types.ListAttribute(["foo", "bar", "baz"])
    assert o.get_for_api() == "foo,bar,baz"


def test_list_attribute_does_not_split_string():
    o = types.ListAttribute("foo")
    assert o.get_for_api() == "foo"


def test_lowercase_string_attribute_get_for_api():
    o = types.LowercaseStringAttribute("FOO")
    assert o.get_for_api() == "foo"



def test_list_instance_variables(gitlab_cli, gl):
    cmd = ["variable", "list"]
    ret = gitlab_cli(cmd)

    assert ret.success


def test_list_group_variables(gitlab_cli, group):
    cmd = ["group-variable", "list", "--group-id", group.id]
    ret = gitlab_cli(cmd)

    assert ret.success


def test_list_project_variables(gitlab_cli, project):
    cmd = ["project-variable", "list", "--project-id", project.id]
    ret = gitlab_cli(cmd)

    assert ret.success



"""
GitLab API: https://docs.gitlab.com/ce/api/todos.html
"""

import json
import os

import pytest
import responses

from gitlab.v4.objects import Todo


with open(os.path.dirname(__file__) + "/../data/todo.json", "r") as json_file:
    todo_content = json_file.read()
    json_content = json.loads(todo_content)


@pytest.fixture
def resp_todo():
    with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/todos",
            json=json_content,
            content_type="application/json",
            status=200,
        )
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/todos/102/mark_as_done",
            json=json_content[0],
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_mark_all_as_done():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/todos/mark_as_done",
            json={},
            content_type="application/json",
            status=204,
        )
        yield rsps


def test_todo(gl, resp_todo):
    todo = gl.todos.list()[0]
    assert isinstance(todo, Todo)
    assert todo.id == 102
    assert todo.target_type == "MergeRequest"
    assert todo.target["assignee"]["username"] == "root"

    todo.mark_as_done()


def test_todo_mark_all_as_done(gl, resp_mark_all_as_done):
    gl.todos.mark_all_as_done()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 Gauvain Pocentek <gauvain@pocentek.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gitlab import utils


def test_clean_str_id():
    src = "nothing_special"
    dest = "nothing_special"
    assert dest == utils.clean_str_id(src)

    src = "foo#bar/baz/"
    dest = "foo%23bar%2Fbaz%2F"
    assert dest == utils.clean_str_id(src)


def test_sanitized_url():
    src = "http://localhost/foo/bar"
    dest = "http://localhost/foo/bar"
    assert dest == utils.sanitized_url(src)

    src = "http://localhost/foo.bar.baz"
    dest = "http://localhost/foo%2Ebar%2Ebaz"
    assert dest == utils.sanitized_url(src)


def test_sanitize_parameters_does_nothing():
    assert 1 == utils.sanitize_parameters(1)
    assert 1.5 == utils.sanitize_parameters(1.5)
    assert "foo" == utils.sanitize_parameters("foo")


def test_sanitize_parameters_slash():
    assert "foo%2Fbar" == utils.sanitize_parameters("foo/bar")


def test_sanitize_parameters_dict():
    source = {"url": "foo/bar", "id": 1}
    expected = {"url": "foo%2Fbar", "id": 1}
    assert expected == utils.sanitize_parameters(source)



"""
GitLab API: https://docs.gitlab.com/ce/api/commits.html
"""

import pytest
import responses


@pytest.fixture
def resp_create_commit():
    content = {
        "id": "ed899a2f4b50b4370feeea94676502b42383c746",
        "short_id": "ed899a2f",
        "title": "Commit message",
    }

    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/projects/1/repository/commits",
            json=content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_commit():
    get_content = {
        "id": "6b2257eabcec3db1f59dafbd84935e3caea04235",
        "short_id": "6b2257ea",
        "title": "Initial commit",
    }
    revert_content = {
        "id": "8b090c1b79a14f2bd9e8a738f717824ff53aebad",
        "short_id": "8b090c1b",
        "title": 'Revert "Initial commit"',
    }

    with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/repository/commits/6b2257ea",
            json=get_content,
            content_type="application/json",
            status=200,
        )
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/projects/1/repository/commits/6b2257ea/revert",
            json=revert_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_get_commit_gpg_signature():
    content = {
        "gpg_key_id": 1,
        "gpg_key_primary_keyid": "8254AAB3FBD54AC9",
        "gpg_key_user_name": "John Doe",
        "gpg_key_user_email": "johndoe@example.com",
        "verification_status": "verified",
        "gpg_key_subkey_id": None,
    }

    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/repository/commits/6b2257ea/signature",
            json=content,
            content_type="application/json",
            status=200,
        )
        yield rsps


def test_get_commit(project, resp_commit):
    commit = project.commits.get("6b2257ea")
    assert commit.short_id == "6b2257ea"
    assert commit.title == "Initial commit"


def test_create_commit(project, resp_create_commit):
    data = {
        "branch": "master",
        "commit_message": "Commit message",
        "actions": [
            {
                "action": "create",
                "file_path": "README",
                "content": "",
            }
        ],
    }
    commit = project.commits.create(data)
    assert commit.short_id == "ed899a2f"
    assert commit.title == data["commit_message"]


def test_revert_commit(project, resp_commit):
    commit = project.commits.get("6b2257ea", lazy=True)
    revert_commit = commit.revert(branch="master")
    assert revert_commit["short_id"] == "8b090c1b"
    assert revert_commit["title"] == 'Revert "Initial commit"'


def test_get_commit_gpg_signature(project, resp_get_commit_gpg_signature):
    commit = project.commits.get("6b2257ea", lazy=True)
    signature = commit.signature()
    assert signature["gpg_key_primary_keyid"] == "8254AAB3FBD54AC9"
    assert signature["verification_status"] == "verified"



"""
GitLab API: https://docs.gitlab.com/ee/api/job_artifacts.html
"""

import pytest
import responses


ref_name = "master"
job = "build"


@pytest.fixture
def resp_artifacts_by_ref_name(binary_content):
    url = f"http://localhost/api/v4/projects/1/jobs/artifacts/{ref_name}/download?job={job}"

    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url=url,
            body=binary_content,
            content_type="application/octet-stream",
            status=200,
        )
        yield rsps


def test_download_artifacts_by_ref_name(gl, binary_content, resp_artifacts_by_ref_name):
    project = gl.projects.get(1, lazy=True)
    artifacts = project.artifacts(ref_name=ref_name, job=job)
    assert artifacts == binary_content



import base64
import os
import time

import requests

import gitlab

LOGIN = "root"
PASSWORD = "5iveL!fe"

SSH_KEY = (
    "ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABAQDZAjAX8vTiHD7Yi3/EzuVaDChtih"
    "79HyJZ6H9dEqxFfmGA1YnncE0xujQ64TCebhkYJKzmTJCImSVkOu9C4hZgsw6eE76n"
    "+Cg3VwEeDUFy+GXlEJWlHaEyc3HWioxgOALbUp3rOezNh+d8BDwwqvENGoePEBsz5l"
    "a6WP5lTi/HJIjAl6Hu+zHgdj1XVExeH+S52EwpZf/ylTJub0Bl5gHwf/siVE48mLMI"
    "sqrukXTZ6Zg+8EHAIvIQwJ1dKcXe8P5IoLT7VKrbkgAnolS0I8J+uH7KtErZJb5oZh"
    "S4OEwsNpaXMAr+6/wWSpircV2/e7sFLlhlKBC4Iq1MpqlZ7G3p foo@bar"
)
DEPLOY_KEY = (
    "ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABAQDFdRyjJQh+1niBpXqE2I8dzjG"
    "MXFHlRjX9yk/UfOn075IdaockdU58sw2Ai1XIWFpZpfJkW7z+P47ZNSqm1gzeXI"
    "rtKa9ZUp8A7SZe8vH4XVn7kh7bwWCUirqtn8El9XdqfkzOs/+FuViriUWoJVpA6"
    "WZsDNaqINFKIA5fj/q8XQw+BcS92L09QJg9oVUuH0VVwNYbU2M2IRmSpybgC/gu"
    "uWTrnCDMmLItksATifLvRZwgdI8dr+q6tbxbZknNcgEPrI2jT0hYN9ZcjNeWuyv"
    "rke9IepE7SPBT41C+YtUX4dfDZDmczM1cE0YL/krdUCfuZHMa4ZS2YyNd6slufc"
    "vn bar@foo"
)

GPG_KEY = """-----BEGIN PGP PUBLIC KEY BLOCK-----

mQENBFn5mzYBCADH6SDVPAp1zh/hxmTi0QplkOfExBACpuY6OhzNdIg+8/528b3g
Y5YFR6T/HLv/PmeHskUj21end1C0PNG2T9dTx+2Vlh9ISsSG1kyF9T5fvMR3bE0x
Dl6S489CXZrjPTS9SHk1kF+7dwjUxLJyxF9hPiSihFefDFu3NeOtG/u8vbC1mewQ
ZyAYue+mqtqcCIFFoBz7wHKMWjIVSJSyTkXExu4OzpVvy3l2EikbvavI3qNz84b+
Mgkv/kiBlNoCy3CVuPk99RYKZ3lX1vVtqQ0OgNGQvb4DjcpyjmbKyibuZwhDjIOh
au6d1OyEbayTntd+dQ4j9EMSnEvm/0MJ4eXPABEBAAG0G0dpdGxhYlRlc3QxIDxm
YWtlQGZha2UudGxkPokBNwQTAQgAIQUCWfmbNgIbAwULCQgHAgYVCAkKCwIEFgID
AQIeAQIXgAAKCRBgxELHf8f3hF3yB/wNJlWPKY65UsB4Lo0hs1OxdxCDqXogSi0u
6crDEIiyOte62pNZKzWy8TJcGZvznRTZ7t8hXgKFLz3PRMcl+vAiRC6quIDUj+2V
eYfwaItd1lUfzvdCaC7Venf4TQ74f5vvNg/zoGwE6eRoSbjlLv9nqsxeA0rUBUQL
LYikWhVMP3TrlfgfduYvh6mfgh57BDLJ9kJVpyfxxx9YLKZbaas9sPa6LgBtR555
JziUxHmbEv8XCsUU8uoFeP1pImbNBplqE3wzJwzOMSmmch7iZzrAwfN7N2j3Wj0H
B5kQddJ9dmB4BbU0IXGhWczvdpxboI2wdY8a1JypxOdePoph/43iuQENBFn5mzYB
CADnTPY0Zf3d9zLjBNgIb3yDl94uOcKCq0twNmyjMhHzGqw+UMe9BScy34GL94Al
xFRQoaL+7P8hGsnsNku29A/VDZivcI+uxTx4WQ7OLcn7V0bnHV4d76iky2ufbUt/
GofthjDs1SonePO2N09sS4V4uK0d5N4BfCzzXgvg8etCLxNmC9BGt7AaKUUzKBO4
2QvNNaC2C/8XEnOgNWYvR36ylAXAmo0sGFXUsBCTiq1fugS9pwtaS2JmaVpZZ3YT
pMZlS0+SjC5BZYFqSmKCsA58oBRzCxQz57nR4h5VEflgD+Hy0HdW0UHETwz83E6/
U0LL6YyvhwFr6KPq5GxinSvfABEBAAGJAR8EGAEIAAkFAln5mzYCGwwACgkQYMRC
x3/H94SJgwgAlKQb10/xcL/epdDkR7vbiei7huGLBpRDb/L5fM8B5W77Qi8Xmuqj
cCu1j99ZCA5hs/vwVn8j8iLSBGMC5gxcuaar/wtmiaEvT9fO/h6q4opG7NcuiJ8H
wRj8ccJmRssNqDD913PLz7T40Ts62blhrEAlJozGVG/q7T3RAZcskOUHKeHfc2RI
YzGsC/I9d7k6uxAv1L9Nm5F2HaAQDzhkdd16nKkGaPGR35cT1JLInkfl5cdm7ldN
nxs4TLO3kZjUTgWKdhpgRNF5hwaz51ZjpebaRf/ZqRuNyX4lIRolDxzOn/+O1o8L
qG2ZdhHHmSK2LaQLFiSprUkikStNU9BqSQ==
=5OGa
-----END PGP PUBLIC KEY BLOCK-----"""
AVATAR_PATH = os.path.join(os.path.dirname(__file__), "avatar.png")


# token authentication from config file
gl = gitlab.Gitlab.from_config(config_files=["/tmp/python-gitlab.cfg"])
gl.auth()
assert isinstance(gl.user, gitlab.v4.objects.CurrentUser)

# markdown
html = gl.markdown("foo")
assert "foo" in html

success, errors = gl.lint("Invalid")
assert success is False
assert errors

# sidekiq
out = gl.sidekiq.queue_metrics()
assert isinstance(out, dict)
assert "pages" in out["queues"]
out = gl.sidekiq.process_metrics()
assert isinstance(out, dict)
assert "hostname" in out["processes"][0]
out = gl.sidekiq.job_stats()
assert isinstance(out, dict)
assert "processed" in out["jobs"]
out = gl.sidekiq.compound_metrics()
assert isinstance(out, dict)
assert "jobs" in out
assert "processes" in out
assert "queues" in out

# settings
settings = gl.settings.get()
settings.default_projects_limit = 42
settings.save()
settings = gl.settings.get()
assert settings.default_projects_limit == 42

# users
new_user = gl.users.create(
    {
        "email": "foo@bar.com",
        "username": "foo",
        "name": "foo",
        "password": "foo_password",
        "avatar": open(AVATAR_PATH, "rb"),
    }
)
avatar_url = new_user.avatar_url.replace("gitlab.test", "localhost:8080")
uploaded_avatar = requests.get(avatar_url).content
assert uploaded_avatar == open(AVATAR_PATH, "rb").read()
users_list = gl.users.list()
for user in users_list:
    if user.username == "foo":
        break
assert new_user.username == user.username
assert new_user.email == user.email

new_user.block()
new_user.unblock()

# user projects list
assert len(new_user.projects.list()) == 0

# events list
new_user.events.list()

foobar_user = gl.users.create(
    {
        "email": "foobar@example.com",
        "username": "foobar",
        "name": "Foo Bar",
        "password": "foobar_password",
    }
)

assert gl.users.list(search="foobar")[0].id == foobar_user.id
expected = [new_user, foobar_user]
actual = list(gl.users.list(search="foo"))
assert len(expected) == len(actual)
assert len(gl.users.list(search="asdf")) == 0
foobar_user.bio = "This is the user bio"
foobar_user.save()

# GPG keys
gkey = new_user.gpgkeys.create({"key": GPG_KEY})
assert len(new_user.gpgkeys.list()) == 1
# Seems broken on the gitlab side
# gkey = new_user.gpgkeys.get(gkey.id)
gkey.delete()
assert len(new_user.gpgkeys.list()) == 0

# SSH keys
key = new_user.keys.create({"title": "testkey", "key": SSH_KEY})
assert len(new_user.keys.list()) == 1
key.delete()
assert len(new_user.keys.list()) == 0

# emails
email = new_user.emails.create({"email": "foo2@bar.com"})
assert len(new_user.emails.list()) == 1
email.delete()
assert len(new_user.emails.list()) == 0

# custom attributes
attrs = new_user.customattributes.list()
assert len(attrs) == 0
attr = new_user.customattributes.set("key", "value1")
assert len(gl.users.list(custom_attributes={"key": "value1"})) == 1
assert attr.key == "key"
assert attr.value == "value1"
assert len(new_user.customattributes.list()) == 1
attr = new_user.customattributes.set("key", "value2")
attr = new_user.customattributes.get("key")
assert attr.value == "value2"
assert len(new_user.customattributes.list()) == 1
attr.delete()
assert len(new_user.customattributes.list()) == 0

# impersonation tokens
user_token = new_user.impersonationtokens.create(
    {"name": "token1", "scopes": ["api", "read_user"]}
)
l = new_user.impersonationtokens.list(state="active")
assert len(l) == 1
user_token.delete()
l = new_user.impersonationtokens.list(state="active")
assert len(l) == 0
l = new_user.impersonationtokens.list(state="inactive")
assert len(l) == 1

new_user.delete()
foobar_user.delete()
assert len(gl.users.list()) == 3 + len(
    [u for u in gl.users.list() if u.username == "ghost"]
)

# current user mail
mail = gl.user.emails.create({"email": "current@user.com"})
assert len(gl.user.emails.list()) == 1
mail.delete()
assert len(gl.user.emails.list()) == 0

# current user GPG keys
gkey = gl.user.gpgkeys.create({"key": GPG_KEY})
assert len(gl.user.gpgkeys.list()) == 1
# Seems broken on the gitlab side
gkey = gl.user.gpgkeys.get(gkey.id)
gkey.delete()
assert len(gl.user.gpgkeys.list()) == 0

# current user key
key = gl.user.keys.create({"title": "testkey", "key": SSH_KEY})
assert len(gl.user.keys.list()) == 1
key.delete()
assert len(gl.user.keys.list()) == 0

# templates
assert gl.dockerfiles.list()
dockerfile = gl.dockerfiles.get("Node")
assert dockerfile.content is not None

assert gl.gitignores.list()
gitignore = gl.gitignores.get("Node")
assert gitignore.content is not None

assert gl.gitlabciymls.list()
gitlabciyml = gl.gitlabciymls.get("Nodejs")
assert gitlabciyml.content is not None

assert gl.licenses.list()
license = gl.licenses.get(
    "bsd-2-clause", project="mytestproject", fullname="mytestfullname"
)
assert "mytestfullname" in license.content

# groups
user1 = gl.users.create(
    {
        "email": "user1@test.com",
        "username": "user1",
        "name": "user1",
        "password": "user1_pass",
    }
)
user2 = gl.users.create(
    {
        "email": "user2@test.com",
        "username": "user2",
        "name": "user2",
        "password": "user2_pass",
    }
)
group1 = gl.groups.create({"name": "group1", "path": "group1"})
group2 = gl.groups.create({"name": "group2", "path": "group2"})

p_id = gl.groups.list(search="group2")[0].id
group3 = gl.groups.create({"name": "group3", "path": "group3", "parent_id": p_id})
group4 = gl.groups.create({"name": "group4", "path": "group4"})

assert len(gl.groups.list()) == 4
assert len(gl.groups.list(search="oup1")) == 1
assert group3.parent_id == p_id
assert group2.subgroups.list()[0].id == group3.id

group1.members.create({"access_level": gitlab.const.OWNER_ACCESS, "user_id": user1.id})
group1.members.create({"access_level": gitlab.const.GUEST_ACCESS, "user_id": user2.id})

group2.members.create({"access_level": gitlab.const.OWNER_ACCESS, "user_id": user2.id})

group4.share(group1.id, gitlab.const.DEVELOPER_ACCESS)
group4.share(group2.id, gitlab.const.MAINTAINER_ACCESS)
# Reload group4 to have updated shared_with_groups
group4 = gl.groups.get(group4.id)
assert len(group4.shared_with_groups) == 2
group4.unshare(group1.id)
# Reload group4 to have updated shared_with_groups
group4 = gl.groups.get(group4.id)
assert len(group4.shared_with_groups) == 1

# User memberships (admin only)
memberships1 = user1.memberships.list()
assert len(memberships1) == 1

memberships2 = user2.memberships.list()
assert len(memberships2) == 2

membership = memberships1[0]
assert membership.source_type == "Namespace"
assert membership.access_level == gitlab.const.OWNER_ACCESS

project_memberships = user1.memberships.list(type="Project")
assert len(project_memberships) == 0

group_memberships = user1.memberships.list(type="Namespace")
assert len(group_memberships) == 1

try:
    membership = user1.memberships.list(type="Invalid")
except gitlab.GitlabListError as e:
    error_message = e.error_message
assert error_message == "type does not have a valid value"

try:
    user1.memberships.list(sudo=user1.name)
except gitlab.GitlabListError as e:
    error_message = e.error_message
assert error_message == "403 Forbidden"

# Administrator belongs to the groups
assert len(group1.members.list()) == 3
assert len(group2.members.list()) == 2

group1.members.delete(user1.id)
assert len(group1.members.list()) == 2
assert len(group1.members.all())
member = group1.members.get(user2.id)
member.access_level = gitlab.const.OWNER_ACCESS
member.save()
member = group1.members.get(user2.id)
assert member.access_level == gitlab.const.OWNER_ACCESS

group2.members.delete(gl.user.id)

# group custom attributes
attrs = group2.customattributes.list()
assert len(attrs) == 0
attr = group2.customattributes.set("key", "value1")
assert len(gl.groups.list(custom_attributes={"key": "value1"})) == 1
assert attr.key == "key"
assert attr.value == "value1"
assert len(group2.customattributes.list()) == 1
attr = group2.customattributes.set("key", "value2")
attr = group2.customattributes.get("key")
assert attr.value == "value2"
assert len(group2.customattributes.list()) == 1
attr.delete()
assert len(group2.customattributes.list()) == 0

# group notification settings
settings = group2.notificationsettings.get()
settings.level = "disabled"
settings.save()
settings = group2.notificationsettings.get()
assert settings.level == "disabled"

# group badges
badge_image = "http://example.com"
badge_link = "http://example/img.svg"
badge = group2.badges.create({"link_url": badge_link, "image_url": badge_image})
assert len(group2.badges.list()) == 1
badge.image_url = "http://another.example.com"
badge.save()
badge = group2.badges.get(badge.id)
assert badge.image_url == "http://another.example.com"
badge.delete()
assert len(group2.badges.list()) == 0

# group milestones
gm1 = group1.milestones.create({"title": "groupmilestone1"})
assert len(group1.milestones.list()) == 1
gm1.due_date = "2020-01-01T00:00:00Z"
gm1.save()
gm1.state_event = "close"
gm1.save()
gm1 = group1.milestones.get(gm1.id)
assert gm1.state == "closed"
assert len(gm1.issues()) == 0
assert len(gm1.merge_requests()) == 0


# group labels
# group1.labels.create({"name": "foo", "description": "bar", "color": "#112233"})
# g_l = group1.labels.get("foo")
# assert g_l.description == "bar"
# g_l.description = "baz"
# g_l.save()
# g_l = group1.labels.get("foo")
# assert g_l.description == "baz"
# assert len(group1.labels.list()) == 1
# g_l.delete()
# assert len(group1.labels.list()) == 0


# group import/export
export = group1.exports.create()
assert export.message == "202 Accepted"

# We cannot check for export_status with group export API
time.sleep(10)

import_archive = "/tmp/gitlab-group-export.tgz"
import_path = "imported_group"
import_name = "Imported Group"

with open(import_archive, "wb") as f:
    export.download(streamed=True, action=f.write)

with open(import_archive, "rb") as f:
    output = gl.groups.import_group(f, import_path, import_name)
assert output["message"] == "202 Accepted"

# We cannot check for returned ID with group import API
time.sleep(10)
group_import = gl.groups.get(import_path)

assert group_import.path == import_path
assert group_import.name == import_name


# hooks
hook = gl.hooks.create({"url": "http://whatever.com"})
assert len(gl.hooks.list()) == 1
hook.delete()
assert len(gl.hooks.list()) == 0

# projects
admin_project = gl.projects.create({"name": "admin_project"})
gr1_project = gl.projects.create({"name": "gr1_project", "namespace_id": group1.id})
gr2_project = gl.projects.create({"name": "gr2_project", "namespace_id": group2.id})
sudo_project = gl.projects.create({"name": "sudo_project"}, sudo=user1.name)

assert len(gl.projects.list(owned=True)) == 3
assert len(gl.projects.list(search="admin")) == 1
assert len(gl.projects.list(as_list=False)) == 4

# test pagination
l1 = gl.projects.list(per_page=1, page=1)
l2 = gl.projects.list(per_page=1, page=2)
assert len(l1) == 1
assert len(l2) == 1
assert l1[0].id != l2[0].id

# group custom attributes
attrs = admin_project.customattributes.list()
assert len(attrs) == 0
attr = admin_project.customattributes.set("key", "value1")
assert len(gl.projects.list(custom_attributes={"key": "value1"})) == 1
assert attr.key == "key"
assert attr.value == "value1"
assert len(admin_project.customattributes.list()) == 1
attr = admin_project.customattributes.set("key", "value2")
attr = admin_project.customattributes.get("key")
assert attr.value == "value2"
assert len(admin_project.customattributes.list()) == 1
attr.delete()
assert len(admin_project.customattributes.list()) == 0

# project pages domains
domain = admin_project.pagesdomains.create({"domain": "foo.domain.com"})
assert len(admin_project.pagesdomains.list()) == 1
assert len(gl.pagesdomains.list()) == 1
domain = admin_project.pagesdomains.get("foo.domain.com")
assert domain.domain == "foo.domain.com"
domain.delete()
assert len(admin_project.pagesdomains.list()) == 0

# project content (files)
admin_project.files.create(
    {
        "file_path": "README",
        "branch": "master",
        "content": "Initial content",
        "commit_message": "Initial commit",
    }
)
readme = admin_project.files.get(file_path="README", ref="master")
readme.content = base64.b64encode(b"Improved README").decode()
time.sleep(2)
readme.save(branch="master", commit_message="new commit")
readme.delete(commit_message="Removing README", branch="master")

admin_project.files.create(
    {
        "file_path": "README.rst",
        "branch": "master",
        "content": "Initial content",
        "commit_message": "New commit",
    }
)
readme = admin_project.files.get(file_path="README.rst", ref="master")
# The first decode() is the ProjectFile method, the second one is the bytes
# object method
assert readme.decode().decode() == "Initial content"

blame = admin_project.files.blame(file_path="README.rst", ref="master")

data = {
    "branch": "master",
    "commit_message": "blah blah blah",
    "actions": [{"action": "create", "file_path": "blah", "content": "blah"}],
}
admin_project.commits.create(data)
assert "@@" in admin_project.commits.list()[0].diff()[0]["diff"]

# commit status
commit = admin_project.commits.list()[0]
# size = len(commit.statuses.list())
# status = commit.statuses.create({"state": "success", "sha": commit.id})
# assert len(commit.statuses.list()) == size + 1

# assert commit.refs()
# assert commit.merge_requests()

# commit signature (for unsigned commits)
# TODO: reasonable tests for signed commits?
try:
    signature = commit.signature()
except gitlab.GitlabGetError as e:
    error_message = e.error_message
assert error_message == "404 Signature Not Found"

# commit comment
commit.comments.create({"note": "This is a commit comment"})
# assert len(commit.comments.list()) == 1

# commit discussion
count = len(commit.discussions.list())
discussion = commit.discussions.create({"body": "Discussion body"})
# assert len(commit.discussions.list()) == (count + 1)
d_note = discussion.notes.create({"body": "first note"})
d_note_from_get = discussion.notes.get(d_note.id)
d_note_from_get.body = "updated body"
d_note_from_get.save()
discussion = commit.discussions.get(discussion.id)
# assert discussion.attributes["notes"][-1]["body"] == "updated body"
d_note_from_get.delete()
discussion = commit.discussions.get(discussion.id)
# assert len(discussion.attributes["notes"]) == 1

# Revert commit
revert_commit = commit.revert(branch="master")

expected_message = 'Revert "{}"\n\nThis reverts commit {}'.format(
    commit.message, commit.id
)
assert revert_commit["message"] == expected_message

try:
    commit.revert(branch="master")
    # Only here to really ensure expected error without a full test framework
    raise AssertionError("Two revert attempts should raise GitlabRevertError")
except gitlab.GitlabRevertError:
    pass

# housekeeping
admin_project.housekeeping()

# repository
tree = admin_project.repository_tree()
assert len(tree) != 0
assert tree[0]["name"] == "README.rst"
blob_id = tree[0]["id"]
blob = admin_project.repository_raw_blob(blob_id)
assert blob.decode() == "Initial content"
archive1 = admin_project.repository_archive()
archive2 = admin_project.repository_archive("master")
assert archive1 == archive2
snapshot = admin_project.snapshot()

# project file uploads
filename = "test.txt"
file_contents = "testing contents"
uploaded_file = admin_project.upload(filename, file_contents)
assert uploaded_file["alt"] == filename
assert uploaded_file["url"].startswith("/uploads/")
assert uploaded_file["url"].endswith("/" + filename)
assert uploaded_file["markdown"] == "[{}]({})".format(
    uploaded_file["alt"], uploaded_file["url"]
)

# environments
admin_project.environments.create(
    {"name": "env1", "external_url": "http://fake.env/whatever"}
)
envs = admin_project.environments.list()
assert len(envs) == 1
env = envs[0]
env.external_url = "http://new.env/whatever"
env.save()
env = admin_project.environments.list()[0]
assert env.external_url == "http://new.env/whatever"
env.stop()
env.delete()
assert len(admin_project.environments.list()) == 0

# Project clusters
admin_project.clusters.create(
    {
        "name": "cluster1",
        "platform_kubernetes_attributes": {
            "api_url": "http://url",
            "token": "tokenval",
        },
    }
)
clusters = admin_project.clusters.list()
assert len(clusters) == 1
cluster = clusters[0]
cluster.platform_kubernetes_attributes = {"api_url": "http://newurl"}
cluster.save()
cluster = admin_project.clusters.list()[0]
assert cluster.platform_kubernetes["api_url"] == "http://newurl"
cluster.delete()
assert len(admin_project.clusters.list()) == 0

# Group clusters
group1.clusters.create(
    {
        "name": "cluster1",
        "platform_kubernetes_attributes": {
            "api_url": "http://url",
            "token": "tokenval",
        },
    }
)
clusters = group1.clusters.list()
assert len(clusters) == 1
cluster = clusters[0]
cluster.platform_kubernetes_attributes = {"api_url": "http://newurl"}
cluster.save()
cluster = group1.clusters.list()[0]
assert cluster.platform_kubernetes["api_url"] == "http://newurl"
cluster.delete()
assert len(group1.clusters.list()) == 0

# project events
admin_project.events.list()

# forks
fork = admin_project.forks.create({"namespace": user1.username})
p = gl.projects.get(fork.id)
assert p.forked_from_project["id"] == admin_project.id

forks = admin_project.forks.list()
assert fork.id in map(lambda p: p.id, forks)

# project hooks
hook = admin_project.hooks.create({"url": "http://hook.url"})
assert len(admin_project.hooks.list()) == 1
hook.note_events = True
hook.save()
hook = admin_project.hooks.get(hook.id)
assert hook.note_events is True
hook.delete()

# deploy keys
deploy_key = admin_project.keys.create({"title": "foo@bar", "key": DEPLOY_KEY})
project_keys = list(admin_project.keys.list())
assert len(project_keys) == 1

sudo_project.keys.enable(deploy_key.id)
assert len(sudo_project.keys.list()) == 1
sudo_project.keys.delete(deploy_key.id)
assert len(sudo_project.keys.list()) == 0

# deploy tokens
deploy_token = admin_project.deploytokens.create(
    {
        "name": "foo",
        "username": "bar",
        "expires_at": "2022-01-01",
        "scopes": ["read_registry"],
    }
)
assert len(admin_project.deploytokens.list()) == 1
assert gl.deploytokens.list() == admin_project.deploytokens.list()

assert admin_project.deploytokens.list()[0].name == "foo"
assert admin_project.deploytokens.list()[0].expires_at == "2022-01-01T00:00:00.000Z"
assert admin_project.deploytokens.list()[0].scopes == ["read_registry"]
# Uncomment once https://gitlab.com/gitlab-org/gitlab/-/issues/211963 is fixed
# assert admin_project.deploytokens.list()[0].username == "bar"
deploy_token.delete()
assert len(admin_project.deploytokens.list()) == 0
# Uncomment once https://gitlab.com/gitlab-org/gitlab/-/issues/212523 is fixed
# assert len(gl.deploytokens.list()) == 0


deploy_token_group = gl.groups.create(
    {"name": "deploy_token_group", "path": "deploy_token_group"}
)

# Uncomment once https://gitlab.com/gitlab-org/gitlab/-/issues/211878 is fixed
# deploy_token = group_deploy_token.deploytokens.create(
#     {
#         "name": "foo",
#         "scopes": ["read_registry"],
#     }
# )

# Remove once https://gitlab.com/gitlab-org/gitlab/-/issues/211878 is fixed
deploy_token = deploy_token_group.deploytokens.create(
    {
        "name": "foo",
        "username": "",
        "expires_at": "",
        "scopes": ["read_repository"],
    }
)

assert len(deploy_token_group.deploytokens.list()) == 1
# Uncomment once https://gitlab.com/gitlab-org/gitlab/-/issues/212523 is fixed
# assert gl.deploytokens.list() == deploy_token_group.deploytokens.list()
deploy_token.delete()
assert len(deploy_token_group.deploytokens.list()) == 0
# Uncomment once https://gitlab.com/gitlab-org/gitlab/-/issues/212523 is fixed
# assert len(gl.deploytokens.list()) == 0

deploy_token_group.delete()

# labels
# label1 = admin_project.labels.create({"name": "label1", "color": "#778899"})
# label1 = admin_project.labels.list()[0]
# assert len(admin_project.labels.list()) == 1
# label1.new_name = "label1updated"
# label1.save()
# assert label1.name == "label1updated"
# label1.subscribe()
# assert label1.subscribed == True
# label1.unsubscribe()
# assert label1.subscribed == False
# label1.delete()

# milestones
m1 = admin_project.milestones.create({"title": "milestone1"})
assert len(admin_project.milestones.list()) == 1
m1.due_date = "2020-01-01T00:00:00Z"
m1.save()
m1.state_event = "close"
m1.save()
m1 = admin_project.milestones.get(m1.id)
assert m1.state == "closed"
assert len(m1.issues()) == 0
assert len(m1.merge_requests()) == 0

# issues
issue1 = admin_project.issues.create({"title": "my issue 1", "milestone_id": m1.id})
issue2 = admin_project.issues.create({"title": "my issue 2"})
issue3 = admin_project.issues.create({"title": "my issue 3"})
assert len(admin_project.issues.list()) == 3
issue3.state_event = "close"
issue3.save()
assert len(admin_project.issues.list(state="closed")) == 1
assert len(admin_project.issues.list(state="opened")) == 2
assert len(admin_project.issues.list(milestone="milestone1")) == 1
assert m1.issues().next().title == "my issue 1"
size = len(issue1.notes.list())
note = issue1.notes.create({"body": "This is an issue note"})
assert len(issue1.notes.list()) == size + 1
emoji = note.awardemojis.create({"name": "tractor"})
assert len(note.awardemojis.list()) == 1
emoji.delete()
assert len(note.awardemojis.list()) == 0
note.delete()
assert len(issue1.notes.list()) == size
assert isinstance(issue1.user_agent_detail(), dict)

assert issue1.user_agent_detail()["user_agent"]
assert issue1.participants()
assert type(issue1.closed_by()) == list
assert type(issue1.related_merge_requests()) == list

# issue labels
label2 = admin_project.labels.create({"name": "label2", "color": "#aabbcc"})
issue1.labels = ["label2"]
issue1.save()

assert issue1 in admin_project.issues.list(labels=["label2"])
assert issue1 in admin_project.issues.list(labels="label2")
assert issue1 in admin_project.issues.list(labels="Any")
assert issue1 not in admin_project.issues.list(labels="None")

# issue events
events = issue1.resourcelabelevents.list()
assert events
event = issue1.resourcelabelevents.get(events[0].id)
assert event

# issue milestones
milestones = issue1.resourcemilestoneevents.list()
assert milestones
milestone = issue1.resourcemilestoneevents.get(milestones[0].id)
assert milestone

size = len(issue1.discussions.list())
discussion = issue1.discussions.create({"body": "Discussion body"})
assert len(issue1.discussions.list()) == size + 1
d_note = discussion.notes.create({"body": "first note"})
d_note_from_get = discussion.notes.get(d_note.id)
d_note_from_get.body = "updated body"
d_note_from_get.save()
discussion = issue1.discussions.get(discussion.id)
assert discussion.attributes["notes"][-1]["body"] == "updated body"
d_note_from_get.delete()
discussion = issue1.discussions.get(discussion.id)
assert len(discussion.attributes["notes"]) == 1

# tags
tag1 = admin_project.tags.create({"tag_name": "v1.0", "ref": "master"})
assert len(admin_project.tags.list()) == 1
tag1.set_release_description("Description 1")
tag1.set_release_description("Description 2")
assert tag1.release["description"] == "Description 2"
tag1.delete()

# project snippet
admin_project.snippets_enabled = True
admin_project.save()
snippet = admin_project.snippets.create(
    {
        "title": "snip1",
        "file_name": "foo.py",
        "content": "initial content",
        "visibility": gitlab.v4.objects.VISIBILITY_PRIVATE,
    }
)

assert snippet.user_agent_detail()["user_agent"]

size = len(snippet.discussions.list())
discussion = snippet.discussions.create({"body": "Discussion body"})
assert len(snippet.discussions.list()) == size + 1
d_note = discussion.notes.create({"body": "first note"})
d_note_from_get = discussion.notes.get(d_note.id)
d_note_from_get.body = "updated body"
d_note_from_get.save()
discussion = snippet.discussions.get(discussion.id)
assert discussion.attributes["notes"][-1]["body"] == "updated body"
d_note_from_get.delete()
discussion = snippet.discussions.get(discussion.id)
assert len(discussion.attributes["notes"]) == 1

snippet.file_name = "bar.py"
snippet.save()
snippet = admin_project.snippets.get(snippet.id)
# TO BE RE-ENABLED AFTER 13.1
# assert snippet.content().decode() == "initial content"
assert snippet.file_name == "bar.py"
size = len(admin_project.snippets.list())
snippet.delete()
assert len(admin_project.snippets.list()) == (size - 1)

# triggers
tr1 = admin_project.triggers.create({"description": "trigger1"})
assert len(admin_project.triggers.list()) == 1
tr1.delete()


# branches and merges
to_merge = admin_project.branches.create({"branch": "branch1", "ref": "master"})
admin_project.files.create(
    {
        "file_path": "README2.rst",
        "branch": "branch1",
        "content": "Initial content",
        "commit_message": "New commit in new branch",
    }
)
mr = admin_project.mergerequests.create(
    {"source_branch": "branch1", "target_branch": "master", "title": "MR readme2"}
)

# discussion
size = len(mr.discussions.list())
discussion = mr.discussions.create({"body": "Discussion body"})
assert len(mr.discussions.list()) == size + 1
d_note = discussion.notes.create({"body": "first note"})
d_note_from_get = discussion.notes.get(d_note.id)
d_note_from_get.body = "updated body"
d_note_from_get.save()
discussion = mr.discussions.get(discussion.id)
assert discussion.attributes["notes"][-1]["body"] == "updated body"
d_note_from_get.delete()
discussion = mr.discussions.get(discussion.id)
assert len(discussion.attributes["notes"]) == 1

# mr labels and events
mr.labels = ["label2"]
mr.save()
events = mr.resourcelabelevents.list()
assert events
event = mr.resourcelabelevents.get(events[0].id)
assert event

# mr milestone events
mr.milestone_id = m1.id
mr.save()
milestones = mr.resourcemilestoneevents.list()
assert milestones
milestone = mr.resourcemilestoneevents.get(milestones[0].id)
assert milestone

# rebasing
assert mr.rebase()

# basic testing: only make sure that the methods exist
mr.commits()
mr.changes()
assert mr.participants()

mr.merge()
admin_project.branches.delete("branch1")

try:
    mr.merge()
except gitlab.GitlabMRClosedError:
    pass

# protected branches
p_b = admin_project.protectedbranches.create({"name": "*-stable"})
assert p_b.name == "*-stable"
p_b = admin_project.protectedbranches.get("*-stable")
# master is protected by default when a branch has been created
assert len(admin_project.protectedbranches.list()) == 2
admin_project.protectedbranches.delete("master")
p_b.delete()
assert len(admin_project.protectedbranches.list()) == 0

# stars
admin_project.star()
assert admin_project.star_count == 1
admin_project.unstar()
assert admin_project.star_count == 0

# project boards
# boards = admin_project.boards.list()
# assert(len(boards))
# board = boards[0]
# lists = board.lists.list()
# begin_size = len(lists)
# last_list = lists[-1]
# last_list.position = 0
# last_list.save()
# last_list.delete()
# lists = board.lists.list()
# assert(len(lists) == begin_size - 1)

# project badges
badge_image = "http://example.com"
badge_link = "http://example/img.svg"
badge = admin_project.badges.create({"link_url": badge_link, "image_url": badge_image})
assert len(admin_project.badges.list()) == 1
badge.image_url = "http://another.example.com"
badge.save()
badge = admin_project.badges.get(badge.id)
assert badge.image_url == "http://another.example.com"
badge.delete()
assert len(admin_project.badges.list()) == 0

# project wiki
wiki_content = "Wiki page content"
wp = admin_project.wikis.create({"title": "wikipage", "content": wiki_content})
assert len(admin_project.wikis.list()) == 1
wp = admin_project.wikis.get(wp.slug)
assert wp.content == wiki_content
# update and delete seem broken
# wp.content = 'new content'
# wp.save()
# wp.delete()
# assert(len(admin_project.wikis.list()) == 0)

# namespaces
ns = gl.namespaces.list(all=True)
assert len(ns) != 0
ns = gl.namespaces.list(search="root", all=True)[0]
assert ns.kind == "user"

# features
# Disabled as this fails with GitLab 11.11
# feat = gl.features.set("foo", 30)
# assert feat.name == "foo"
# assert len(gl.features.list()) == 1
# feat.delete()
# assert len(gl.features.list()) == 0

# broadcast messages
msg = gl.broadcastmessages.create({"message": "this is the message"})
msg.color = "#444444"
msg.save()
msg_id = msg.id
msg = gl.broadcastmessages.list(all=True)[0]
assert msg.color == "#444444"
msg = gl.broadcastmessages.get(msg_id)
assert msg.color == "#444444"
msg.delete()
assert len(gl.broadcastmessages.list()) == 0

# notification settings
settings = gl.notificationsettings.get()
settings.level = gitlab.NOTIFICATION_LEVEL_WATCH
settings.save()
settings = gl.notificationsettings.get()
assert settings.level == gitlab.NOTIFICATION_LEVEL_WATCH

# services
service = admin_project.services.get("asana")
service.api_key = "whatever"
service.save()
service = admin_project.services.get("asana")
assert service.active == True
service.delete()
service = admin_project.services.get("asana")
assert service.active == False

# snippets
snippets = gl.snippets.list(all=True)
assert len(snippets) == 0
snippet = gl.snippets.create(
    {"title": "snippet1", "file_name": "snippet1.py", "content": "import gitlab"}
)
snippet = gl.snippets.get(snippet.id)
snippet.title = "updated_title"
snippet.save()
snippet = gl.snippets.get(snippet.id)
assert snippet.title == "updated_title"
content = snippet.content()
assert content.decode() == "import gitlab"

assert snippet.user_agent_detail()["user_agent"]

snippet.delete()
snippets = gl.snippets.list(all=True)
assert len(snippets) == 0

# user activities
gl.user_activities.list(query_parameters={"from": "2019-01-01"})

# events
gl.events.list()

# rate limit
settings = gl.settings.get()
settings.throttle_authenticated_api_enabled = True
settings.throttle_authenticated_api_requests_per_period = 1
settings.throttle_authenticated_api_period_in_seconds = 3
settings.save()
projects = list()
for i in range(0, 20):
    projects.append(gl.projects.create({"name": str(i) + "ok"}))

error_message = None
for i in range(20, 40):
    try:
        projects.append(
            gl.projects.create({"name": str(i) + "shouldfail"}, obey_rate_limit=False)
        )
    except gitlab.GitlabCreateError as e:
        error_message = e.error_message
        break
assert "Retry later" in error_message
settings.throttle_authenticated_api_enabled = False
settings.save()
[current_project.delete() for current_project in projects]

# project import/export
ex = admin_project.exports.create()
ex.refresh()
count = 0
while ex.export_status != "finished":
    time.sleep(1)
    ex.refresh()
    count += 1
    if count == 10:
        raise Exception("Project export taking too much time")
with open("/tmp/gitlab-export.tgz", "wb") as f:
    ex.download(streamed=True, action=f.write)

output = gl.projects.import_project(
    open("/tmp/gitlab-export.tgz", "rb"), "imported_project", name="Imported Project"
)
project_import = gl.projects.get(output["id"], lazy=True).imports.get()

assert project_import.path == "imported_project"
assert project_import.name == "Imported Project"

count = 0
while project_import.import_status != "finished":
    time.sleep(1)
    project_import.refresh()
    count += 1
    if count == 10:
        raise Exception("Project import taking too much time")

# project releases
release_test_project = gl.projects.create(
    {"name": "release-test-project", "initialize_with_readme": True}
)
release_name = "Demo Release"
release_tag_name = "v1.2.3"
release_description = "release notes go here"
release_test_project.releases.create(
    {
        "name": release_name,
        "tag_name": release_tag_name,
        "description": release_description,
        "ref": "master",
    }
)
assert len(release_test_project.releases.list()) == 1

# get single release
retrieved_project = release_test_project.releases.get(release_tag_name)
assert retrieved_project.name == release_name
assert retrieved_project.tag_name == release_tag_name
assert retrieved_project.description == release_description

# delete release
release_test_project.releases.delete(release_tag_name)
assert len(release_test_project.releases.list()) == 0
release_test_project.delete()

# project remote mirrors
mirror_url = "http://gitlab.test/root/mirror.git"

# create remote mirror
mirror = admin_project.remote_mirrors.create({"url": mirror_url})
assert mirror.url == mirror_url

# update remote mirror
mirror.enabled = True
mirror.save()

# list remote mirrors
mirror = admin_project.remote_mirrors.list()[0]
assert isinstance(mirror, gitlab.v4.objects.ProjectRemoteMirror)
assert mirror.url == mirror_url
assert mirror.enabled is True

# status
message = "Test"
emoji = "thumbsup"
status = gl.user.status.get()
status.message = message
status.emoji = emoji
status.save()
new_status = gl.user.status.get()
assert new_status.message == message
assert new_status.emoji == emoji



import pytest
import gitlab


@pytest.fixture
def gl():
    return gitlab.Gitlab(
        "http://localhost",
        private_token="private_token",
        ssl_verify=True,
        api_version=4,
    )


# Todo: parametrize, but check what tests it's really useful for
@pytest.fixture
def gl_trailing():
    return gitlab.Gitlab(
        "http://localhost/", private_token="private_token", api_version=4
    )


@pytest.fixture
def default_config(tmpdir):
    valid_config = """[global]
    default = one
    ssl_verify = true
    timeout = 2

    [one]
    url = http://one.url
    private_token = ABCDEF
    """

    config_path = tmpdir.join("python-gitlab.cfg")
    config_path.write(valid_config)
    return str(config_path)


@pytest.fixture
def group(gl):
    return gl.groups.get(1, lazy=True)


@pytest.fixture
def project(gl):
    return gl.projects.get(1, lazy=True)


@pytest.fixture
def user(gl):
    return gl.users.get(1, lazy=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2016-2017 Gauvain Pocentek <gauvain@pocentek.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import os
import tempfile
import io

from contextlib import redirect_stderr  # noqa: H302

import pytest

from gitlab import cli
import gitlab.v4.cli


def test_what_to_cls():
    assert "Foo" == cli.what_to_cls("foo")
    assert "FooBar" == cli.what_to_cls("foo-bar")


def test_cls_to_what():
    class Class(object):
        pass

    class TestClass(object):
        pass

    assert "test-class" == cli.cls_to_what(TestClass)
    assert "class" == cli.cls_to_what(Class)


def test_die():
    fl = io.StringIO()
    with redirect_stderr(fl):
        with pytest.raises(SystemExit) as test:
            cli.die("foobar")
    assert fl.getvalue() == "foobar\n"
    assert test.value.code == 1


def test_parse_value():
    ret = cli._parse_value("foobar")
    assert ret == "foobar"

    ret = cli._parse_value(True)
    assert ret is True

    ret = cli._parse_value(1)
    assert ret == 1

    ret = cli._parse_value(None)
    assert ret is None

    fd, temp_path = tempfile.mkstemp()
    os.write(fd, b"content")
    os.close(fd)
    ret = cli._parse_value("@%s" % temp_path)
    assert ret == "content"
    os.unlink(temp_path)

    fl = io.StringIO()
    with redirect_stderr(fl):
        with pytest.raises(SystemExit) as exc:
            cli._parse_value("@/thisfileprobablydoesntexist")
        assert (
            fl.getvalue() == "[Errno 2] No such file or directory:"
            " '/thisfileprobablydoesntexist'\n"
        )
        assert exc.value.code == 1


def test_base_parser():
    parser = cli._get_base_parser()
    args = parser.parse_args(["-v", "-g", "gl_id", "-c", "foo.cfg", "-c", "bar.cfg"])
    assert args.verbose
    assert args.gitlab == "gl_id"
    assert args.config_file == ["foo.cfg", "bar.cfg"]


def test_v4_parse_args():
    parser = cli._get_parser(gitlab.v4.cli)
    args = parser.parse_args(["project", "list"])
    assert args.what == "project"
    assert args.whaction == "list"


def test_v4_parser():
    parser = cli._get_parser(gitlab.v4.cli)
    subparsers = next(
        action
        for action in parser._actions
        if isinstance(action, argparse._SubParsersAction)
    )
    assert subparsers is not None
    assert "project" in subparsers.choices

    user_subparsers = next(
        action
        for action in subparsers.choices["project"]._actions
        if isinstance(action, argparse._SubParsersAction)
    )
    assert user_subparsers is not None
    assert "list" in user_subparsers.choices
    assert "get" in user_subparsers.choices
    assert "delete" in user_subparsers.choices
    assert "update" in user_subparsers.choices
    assert "create" in user_subparsers.choices
    assert "archive" in user_subparsers.choices
    assert "unarchive" in user_subparsers.choices

    actions = user_subparsers.choices["create"]._option_string_actions
    assert not actions["--description"].required

    user_subparsers = next(
        action
        for action in subparsers.choices["group"]._actions
        if isinstance(action, argparse._SubParsersAction)
    )
    actions = user_subparsers.choices["create"]._option_string_actions
    assert actions["--name"].required



"""
GitLab API: https://docs.gitlab.com/ce/api/project_snippets.html
             https://docs.gitlab.com/ee/api/snippets.html (todo)
"""

import pytest
import responses


title = "Example Snippet Title"
visibility = "private"
new_title = "new-title"


@pytest.fixture
def resp_snippet():
    content = {
        "title": title,
        "description": "More verbose snippet description",
        "file_name": "example.txt",
        "content": "source code with multiple lines",
        "visibility": visibility,
    }

    with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/snippets",
            json=[content],
            content_type="application/json",
            status=200,
        )
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/snippets/1",
            json=content,
            content_type="application/json",
            status=200,
        )
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/projects/1/snippets",
            json=content,
            content_type="application/json",
            status=200,
        )

        updated_content = dict(content)
        updated_content["title"] = new_title
        updated_content["visibility"] = visibility

        rsps.add(
            method=responses.PUT,
            url="http://localhost/api/v4/projects/1/snippets",
            json=updated_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


def test_list_project_snippets(project, resp_snippet):
    snippets = project.snippets.list()
    assert len(snippets) == 1
    assert snippets[0].title == title
    assert snippets[0].visibility == visibility


def test_get_project_snippet(project, resp_snippet):
    snippet = project.snippets.get(1)
    assert snippet.title == title
    assert snippet.visibility == visibility


def test_create_update_project_snippets(project, resp_snippet):
    snippet = project.snippets.create(
        {
            "title": title,
            "file_name": title,
            "content": title,
            "visibility": visibility,
        }
    )
    assert snippet.title == title
    assert snippet.visibility == visibility

    snippet.title = new_title
    snippet.save()
    assert snippet.title == new_title
    assert snippet.visibility == visibility



import pytest

from httmock import HTTMock, urlmatch, response

from gitlab import *


def test_build_url(gl):
    r = gl._build_url("http://localhost/api/v4")
    assert r == "http://localhost/api/v4"
    r = gl._build_url("https://localhost/api/v4")
    assert r == "https://localhost/api/v4"
    r = gl._build_url("/projects")
    assert r == "http://localhost/api/v4/projects"


def test_http_request(gl):
    @urlmatch(scheme="http", netloc="localhost", path="/api/v4/projects", method="get")
    def resp_cont(url, request):
        headers = {"content-type": "application/json"}
        content = '[{"name": "project1"}]'
        return response(200, content, headers, None, 5, request)

    with HTTMock(resp_cont):
        http_r = gl.http_request("get", "/projects")
        http_r.json()
        assert http_r.status_code == 200


def test_http_request_404(gl):
    @urlmatch(scheme="http", netloc="localhost", path="/api/v4/not_there", method="get")
    def resp_cont(url, request):
        content = {"Here is wh it failed"}
        return response(404, content, {}, None, 5, request)

    with HTTMock(resp_cont):
        with pytest.raises(GitlabHttpError):
            gl.http_request("get", "/not_there")


def test_get_request(gl):
    @urlmatch(scheme="http", netloc="localhost", path="/api/v4/projects", method="get")
    def resp_cont(url, request):
        headers = {"content-type": "application/json"}
        content = '{"name": "project1"}'
        return response(200, content, headers, None, 5, request)

    with HTTMock(resp_cont):
        result = gl.http_get("/projects")
        assert isinstance(result, dict)
        assert result["name"] == "project1"


def test_get_request_raw(gl):
    @urlmatch(scheme="http", netloc="localhost", path="/api/v4/projects", method="get")
    def resp_cont(url, request):
        headers = {"content-type": "application/octet-stream"}
        content = "content"
        return response(200, content, headers, None, 5, request)

    with HTTMock(resp_cont):
        result = gl.http_get("/projects")
        assert result.content.decode("utf-8") == "content"


def test_get_request_404(gl):
    @urlmatch(scheme="http", netloc="localhost", path="/api/v4/not_there", method="get")
    def resp_cont(url, request):
        content = {"Here is wh it failed"}
        return response(404, content, {}, None, 5, request)

    with HTTMock(resp_cont):
        with pytest.raises(GitlabHttpError):
            gl.http_get("/not_there")


def test_get_request_invalid_data(gl):
    @urlmatch(scheme="http", netloc="localhost", path="/api/v4/projects", method="get")
    def resp_cont(url, request):
        headers = {"content-type": "application/json"}
        content = '["name": "project1"]'
        return response(200, content, headers, None, 5, request)

    with HTTMock(resp_cont):
        with pytest.raises(GitlabParsingError):
            gl.http_get("/projects")


def test_list_request(gl):
    @urlmatch(scheme="http", netloc="localhost", path="/api/v4/projects", method="get")
    def resp_cont(url, request):
        headers = {"content-type": "application/json", "X-Total": 1}
        content = '[{"name": "project1"}]'
        return response(200, content, headers, None, 5, request)

    with HTTMock(resp_cont):
        result = gl.http_list("/projects", as_list=True)
        assert isinstance(result, list)
        assert len(result) == 1

    with HTTMock(resp_cont):
        result = gl.http_list("/projects", as_list=False)
        assert isinstance(result, GitlabList)
        assert len(result) == 1

    with HTTMock(resp_cont):
        result = gl.http_list("/projects", all=True)
        assert isinstance(result, list)
        assert len(result) == 1


def test_list_request_404(gl):
    @urlmatch(scheme="http", netloc="localhost", path="/api/v4/not_there", method="get")
    def resp_cont(url, request):
        content = {"Here is why it failed"}
        return response(404, content, {}, None, 5, request)

    with HTTMock(resp_cont):
        with pytest.raises(GitlabHttpError):
            gl.http_list("/not_there")


def test_list_request_invalid_data(gl):
    @urlmatch(scheme="http", netloc="localhost", path="/api/v4/projects", method="get")
    def resp_cont(url, request):
        headers = {"content-type": "application/json"}
        content = '["name": "project1"]'
        return response(200, content, headers, None, 5, request)

    with HTTMock(resp_cont):
        with pytest.raises(GitlabParsingError):
            gl.http_list("/projects")


def test_post_request(gl):
    @urlmatch(scheme="http", netloc="localhost", path="/api/v4/projects", method="post")
    def resp_cont(url, request):
        headers = {"content-type": "application/json"}
        content = '{"name": "project1"}'
        return response(200, content, headers, None, 5, request)

    with HTTMock(resp_cont):
        result = gl.http_post("/projects")
        assert isinstance(result, dict)
        assert result["name"] == "project1"


def test_post_request_404(gl):
    @urlmatch(
        scheme="http", netloc="localhost", path="/api/v4/not_there", method="post"
    )
    def resp_cont(url, request):
        content = {"Here is wh it failed"}
        return response(404, content, {}, None, 5, request)

    with HTTMock(resp_cont):
        with pytest.raises(GitlabHttpError):
            gl.http_post("/not_there")


def test_post_request_invalid_data(gl):
    @urlmatch(scheme="http", netloc="localhost", path="/api/v4/projects", method="post")
    def resp_cont(url, request):
        headers = {"content-type": "application/json"}
        content = '["name": "project1"]'
        return response(200, content, headers, None, 5, request)

    with HTTMock(resp_cont):
        with pytest.raises(GitlabParsingError):
            gl.http_post("/projects")


def test_put_request(gl):
    @urlmatch(scheme="http", netloc="localhost", path="/api/v4/projects", method="put")
    def resp_cont(url, request):
        headers = {"content-type": "application/json"}
        content = '{"name": "project1"}'
        return response(200, content, headers, None, 5, request)

    with HTTMock(resp_cont):
        result = gl.http_put("/projects")
        assert isinstance(result, dict)
        assert result["name"] == "project1"


def test_put_request_404(gl):
    @urlmatch(scheme="http", netloc="localhost", path="/api/v4/not_there", method="put")
    def resp_cont(url, request):
        content = {"Here is wh it failed"}
        return response(404, content, {}, None, 5, request)

    with HTTMock(resp_cont):
        with pytest.raises(GitlabHttpError):
            gl.http_put("/not_there")


def test_put_request_invalid_data(gl):
    @urlmatch(scheme="http", netloc="localhost", path="/api/v4/projects", method="put")
    def resp_cont(url, request):
        headers = {"content-type": "application/json"}
        content = '["name": "project1"]'
        return response(200, content, headers, None, 5, request)

    with HTTMock(resp_cont):
        with pytest.raises(GitlabParsingError):
            gl.http_put("/projects")


def test_delete_request(gl):
    @urlmatch(
        scheme="http", netloc="localhost", path="/api/v4/projects", method="delete"
    )
    def resp_cont(url, request):
        headers = {"content-type": "application/json"}
        content = "true"
        return response(200, content, headers, None, 5, request)

    with HTTMock(resp_cont):
        result = gl.http_delete("/projects")
        assert isinstance(result, requests.Response)
        assert result.json() == True


def test_delete_request_404(gl):
    @urlmatch(
        scheme="http", netloc="localhost", path="/api/v4/not_there", method="delete"
    )
    def resp_cont(url, request):
        content = {"Here is wh it failed"}
        return response(404, content, {}, None, 5, request)

    with HTTMock(resp_cont):
        with pytest.raises(GitlabHttpError):
            gl.http_delete("/not_there")



"""
GitLab API: https://docs.gitlab.com/ce/api/packages.html
"""
import re

import pytest
import responses

from gitlab.v4.objects import GroupPackage, ProjectPackage


package_content = {
    "id": 1,
    "name": "com/mycompany/my-app",
    "version": "1.0-SNAPSHOT",
    "package_type": "maven",
    "_links": {
        "web_path": "/namespace1/project1/-/packages/1",
        "delete_api_path": "/namespace1/project1/-/packages/1",
    },
    "created_at": "2019-11-27T03:37:38.711Z",
    "pipeline": {
        "id": 123,
        "status": "pending",
        "ref": "new-pipeline",
        "sha": "a91957a858320c0e17f3a0eca7cfacbff50ea29a",
        "web_url": "https://example.com/foo/bar/pipelines/47",
        "created_at": "2016-08-11T11:28:34.085Z",
        "updated_at": "2016-08-11T11:32:35.169Z",
        "user": {
            "name": "Administrator",
            "avatar_url": "https://www.gravatar.com/avatar/e64c7d89f26bd1972efa854d13d7dd61?s=80&d=identicon",
        },
    },
    "versions": [
        {
            "id": 2,
            "version": "2.0-SNAPSHOT",
            "created_at": "2020-04-28T04:42:11.573Z",
            "pipeline": {
                "id": 234,
                "status": "pending",
                "ref": "new-pipeline",
                "sha": "a91957a858320c0e17f3a0eca7cfacbff50ea29a",
                "web_url": "https://example.com/foo/bar/pipelines/58",
                "created_at": "2016-08-11T11:28:34.085Z",
                "updated_at": "2016-08-11T11:32:35.169Z",
                "user": {
                    "name": "Administrator",
                    "avatar_url": "https://www.gravatar.com/avatar/e64c7d89f26bd1972efa854d13d7dd61?s=80&d=identicon",
                },
            },
        }
    ],
}


@pytest.fixture
def resp_list_packages():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url=re.compile(r"http://localhost/api/v4/(groups|projects)/1/packages"),
            json=[package_content],
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_get_package():
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/packages/1",
            json=package_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_delete_package(no_content):
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.DELETE,
            url="http://localhost/api/v4/projects/1/packages/1",
            json=no_content,
            content_type="application/json",
            status=204,
        )
        yield rsps


def test_list_project_packages(project, resp_list_packages):
    packages = project.packages.list()
    assert isinstance(packages, list)
    assert isinstance(packages[0], ProjectPackage)
    assert packages[0].version == "1.0-SNAPSHOT"


def test_list_group_packages(group, resp_list_packages):
    packages = group.packages.list()
    assert isinstance(packages, list)
    assert isinstance(packages[0], GroupPackage)
    assert packages[0].version == "1.0-SNAPSHOT"


def test_get_project_package(project, resp_get_package):
    package = project.packages.get(1)
    assert isinstance(package, ProjectPackage)
    assert package.version == "1.0-SNAPSHOT"


def test_delete_project_package(project, resp_delete_package):
    package = project.packages.get(1, lazy=True)
    package.delete()



"""
GitLab API: https://docs.gitlab.com/ce/api/environments.html
"""
import pytest
import responses

from gitlab.v4.objects import ProjectEnvironment


@pytest.fixture
def resp_get_environment():
    content = {"name": "environment_name", "id": 1, "last_deployment": "sometime"}

    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/environments/1",
            json=content,
            content_type="application/json",
            status=200,
        )
        yield rsps


def test_project_environments(project, resp_get_environment):
    environment = project.environments.get(1)
    assert isinstance(environment, ProjectEnvironment)
    assert environment.id == 1
    assert environment.last_deployment == "sometime"
    assert environment.name == "environment_name"



# -*- coding: utf-8 -*-
#
# Copyright (C) 2017 Gauvain Pocentek <gauvain@pocentek.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pickle

from gitlab import base
import pytest


class FakeGitlab(object):
    pass


class FakeObject(base.RESTObject):
    pass


class FakeManager(base.RESTManager):
    _obj_cls = FakeObject
    _path = "/tests"


@pytest.fixture
def fake_gitlab():
    return FakeGitlab()


@pytest.fixture
def fake_manager(fake_gitlab):
    return FakeManager(fake_gitlab)


class TestRESTManager:
    def test_computed_path_simple(self):
        class MGR(base.RESTManager):
            _path = "/tests"
            _obj_cls = object

        mgr = MGR(FakeGitlab())
        assert mgr._computed_path == "/tests"

    def test_computed_path_with_parent(self):
        class MGR(base.RESTManager):
            _path = "/tests/%(test_id)s/cases"
            _obj_cls = object
            _from_parent_attrs = {"test_id": "id"}

        class Parent(object):
            id = 42

        mgr = MGR(FakeGitlab(), parent=Parent())
        assert mgr._computed_path == "/tests/42/cases"

    def test_path_property(self):
        class MGR(base.RESTManager):
            _path = "/tests"
            _obj_cls = object

        mgr = MGR(FakeGitlab())
        assert mgr.path == "/tests"


class TestRESTObject:
    def test_instantiate(self, fake_gitlab, fake_manager):
        obj = FakeObject(fake_manager, {"foo": "bar"})

        assert {"foo": "bar"} == obj._attrs
        assert {} == obj._updated_attrs
        assert None == obj._create_managers()
        assert fake_manager == obj.manager
        assert fake_gitlab == obj.manager.gitlab

    def test_picklability(self, fake_manager):
        obj = FakeObject(fake_manager, {"foo": "bar"})
        original_obj_module = obj._module
        pickled = pickle.dumps(obj)
        unpickled = pickle.loads(pickled)
        assert isinstance(unpickled, FakeObject)
        assert hasattr(unpickled, "_module")
        assert unpickled._module == original_obj_module
        pickled2 = pickle.dumps(unpickled)

    def test_attrs(self, fake_manager):
        obj = FakeObject(fake_manager, {"foo": "bar"})

        assert "bar" == obj.foo
        with pytest.raises(AttributeError):
            getattr(obj, "bar")

        obj.bar = "baz"
        assert "baz" == obj.bar
        assert {"foo": "bar"} == obj._attrs
        assert {"bar": "baz"} == obj._updated_attrs

    def test_get_id(self, fake_manager):
        obj = FakeObject(fake_manager, {"foo": "bar"})
        obj.id = 42
        assert 42 == obj.get_id()

        obj.id = None
        assert None == obj.get_id()

    def test_custom_id_attr(self, fake_manager):
        class OtherFakeObject(FakeObject):
            _id_attr = "foo"

        obj = OtherFakeObject(fake_manager, {"foo": "bar"})
        assert "bar" == obj.get_id()

    def test_update_attrs(self, fake_manager):
        obj = FakeObject(fake_manager, {"foo": "bar"})
        obj.bar = "baz"
        obj._update_attrs({"foo": "foo", "bar": "bar"})
        assert {"foo": "foo", "bar": "bar"} == obj._attrs
        assert {} == obj._updated_attrs

    def test_create_managers(self, fake_gitlab, fake_manager):
        class ObjectWithManager(FakeObject):
            _managers = (("fakes", "FakeManager"),)

        obj = ObjectWithManager(fake_manager, {"foo": "bar"})
        obj.id = 42
        assert isinstance(obj.fakes, FakeManager)
        assert obj.fakes.gitlab == fake_gitlab
        assert obj.fakes._parent == obj

    def test_equality(self, fake_manager):
        obj1 = FakeObject(fake_manager, {"id": "foo"})
        obj2 = FakeObject(fake_manager, {"id": "foo", "other_attr": "bar"})
        assert obj1 == obj2

    def test_equality_custom_id(self, fake_manager):
        class OtherFakeObject(FakeObject):
            _id_attr = "foo"

        obj1 = OtherFakeObject(fake_manager, {"foo": "bar"})
        obj2 = OtherFakeObject(fake_manager, {"foo": "bar", "other_attr": "baz"})
        assert obj1 == obj2

    def test_inequality(self, fake_manager):
        obj1 = FakeObject(fake_manager, {"id": "foo"})
        obj2 = FakeObject(fake_manager, {"id": "bar"})
        assert obj1 != obj2

    def test_inequality_no_id(self, fake_manager):
        obj1 = FakeObject(fake_manager, {"attr1": "foo"})
        obj2 = FakeObject(fake_manager, {"attr1": "bar"})
        assert obj1 != obj2



# -*- coding: utf-8 -*-
#
# Copyright (C) 2016-2017 Gauvain Pocentek <gauvain@pocentek.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

NO_ACCESS = 0
GUEST_ACCESS = 10
REPORTER_ACCESS = 20
DEVELOPER_ACCESS = 30
MAINTAINER_ACCESS = 40
MASTER_ACCESS = MAINTAINER_ACCESS
OWNER_ACCESS = 50

VISIBILITY_PRIVATE = 0
VISIBILITY_INTERNAL = 10
VISIBILITY_PUBLIC = 20

NOTIFICATION_LEVEL_DISABLED = "disabled"
NOTIFICATION_LEVEL_PARTICIPATING = "participating"
NOTIFICATION_LEVEL_WATCH = "watch"
NOTIFICATION_LEVEL_GLOBAL = "global"
NOTIFICATION_LEVEL_MENTION = "mention"
NOTIFICATION_LEVEL_CUSTOM = "custom"



"""
GitLab API: https://docs.gitlab.com/ce/api/users.html
"""
import pytest
import responses

from gitlab.v4.objects import User, UserMembership, UserStatus


@pytest.fixture
def resp_get_user():
    content = {
        "name": "name",
        "id": 1,
        "password": "password",
        "username": "username",
        "email": "email",
    }

    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/users/1",
            json=content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_get_user_memberships():
    content = [
        {
            "source_id": 1,
            "source_name": "Project one",
            "source_type": "Project",
            "access_level": "20",
        },
        {
            "source_id": 3,
            "source_name": "Group three",
            "source_type": "Namespace",
            "access_level": "20",
        },
    ]

    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/users/1/memberships",
            json=content,
            content_type="application/json",
            status=200,
        )
        yield rsps


@pytest.fixture
def resp_activate():
    with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/users/1/activate",
            json={},
            content_type="application/json",
            status=201,
        )
        rsps.add(
            method=responses.POST,
            url="http://localhost/api/v4/users/1/deactivate",
            json={},
            content_type="application/json",
            status=201,
        )
        yield rsps


@pytest.fixture
def resp_get_user_status():
    content = {
        "message": "test",
        "message_html": "<h1>Message</h1>",
        "emoji": "thumbsup",
    }

    with responses.RequestsMock() as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/users/1/status",
            json=content,
            content_type="application/json",
            status=200,
        )
        yield rsps


def test_get_user(gl, resp_get_user):
    user = gl.users.get(1)
    assert isinstance(user, User)
    assert user.name == "name"
    assert user.id == 1


def test_user_memberships(user, resp_get_user_memberships):
    memberships = user.memberships.list()
    assert isinstance(memberships[0], UserMembership)
    assert memberships[0].source_type == "Project"


def test_user_status(user, resp_get_user_status):
    status = user.status.get()
    assert isinstance(status, UserStatus)
    assert status.message == "test"
    assert status.emoji == "thumbsup"


def test_user_activate_deactivate(user, resp_activate):
    user.activate()
    user.deactivate()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2013-2017 Gauvain Pocentek <gauvain@pocentek.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import configparser


def _env_config():
    if "PYTHON_GITLAB_CFG" in os.environ:
        return [os.environ["PYTHON_GITLAB_CFG"]]
    return []


_DEFAULT_FILES = _env_config() + [
    "/etc/python-gitlab.cfg",
    os.path.expanduser("~/.python-gitlab.cfg"),
]


class ConfigError(Exception):
    pass


class GitlabIDError(ConfigError):
    pass


class GitlabDataError(ConfigError):
    pass


class GitlabConfigMissingError(ConfigError):
    pass


class GitlabConfigParser(object):
    def __init__(self, gitlab_id=None, config_files=None):
        self.gitlab_id = gitlab_id
        _files = config_files or _DEFAULT_FILES
        file_exist = False
        for file in _files:
            if os.path.exists(file):
                file_exist = True
        if not file_exist:
            raise GitlabConfigMissingError(
                "Config file not found. \nPlease create one in "
                "one of the following locations: {} \nor "
                "specify a config file using the '-c' parameter.".format(
                    ", ".join(_DEFAULT_FILES)
                )
            )

        self._config = configparser.ConfigParser()
        self._config.read(_files)

        if self.gitlab_id is None:
            try:
                self.gitlab_id = self._config.get("global", "default")
            except Exception as e:
                raise GitlabIDError(
                    "Impossible to get the gitlab id (not specified in config file)"
                ) from e

        try:
            self.url = self._config.get(self.gitlab_id, "url")
        except Exception as e:
            raise GitlabDataError(
                "Impossible to get gitlab informations from "
                "configuration (%s)" % self.gitlab_id
            ) from e

        self.ssl_verify = True
        try:
            self.ssl_verify = self._config.getboolean("global", "ssl_verify")
        except ValueError:
            # Value Error means the option exists but isn't a boolean.
            # Get as a string instead as it should then be a local path to a
            # CA bundle.
            try:
                self.ssl_verify = self._config.get("global", "ssl_verify")
            except Exception:
                pass
        except Exception:
            pass
        try:
            self.ssl_verify = self._config.getboolean(self.gitlab_id, "ssl_verify")
        except ValueError:
            # Value Error means the option exists but isn't a boolean.
            # Get as a string instead as it should then be a local path to a
            # CA bundle.
            try:
                self.ssl_verify = self._config.get(self.gitlab_id, "ssl_verify")
            except Exception:
                pass
        except Exception:
            pass

        self.timeout = 60
        try:
            self.timeout = self._config.getint("global", "timeout")
        except Exception:
            pass
        try:
            self.timeout = self._config.getint(self.gitlab_id, "timeout")
        except Exception:
            pass

        self.private_token = None
        try:
            self.private_token = self._config.get(self.gitlab_id, "private_token")
        except Exception:
            pass

        self.oauth_token = None
        try:
            self.oauth_token = self._config.get(self.gitlab_id, "oauth_token")
        except Exception:
            pass

        self.job_token = None
        try:
            self.job_token = self._config.get(self.gitlab_id, "job_token")
        except Exception:
            pass

        self.http_username = None
        self.http_password = None
        try:
            self.http_username = self._config.get(self.gitlab_id, "http_username")
            self.http_password = self._config.get(self.gitlab_id, "http_password")
        except Exception:
            pass

        self.api_version = "4"
        try:
            self.api_version = self._config.get("global", "api_version")
        except Exception:
            pass
        try:
            self.api_version = self._config.get(self.gitlab_id, "api_version")
        except Exception:
            pass
        if self.api_version not in ("4",):
            raise GitlabDataError("Unsupported API version: %s" % self.api_version)

        self.per_page = None
        for section in ["global", self.gitlab_id]:
            try:
                self.per_page = self._config.getint(section, "per_page")
            except Exception:
                pass
        if self.per_page is not None and not 0 <= self.per_page <= 100:
            raise GitlabDataError("Unsupported per_page number: %s" % self.per_page)

        self.pagination = None
        try:
            self.pagination = self._config.get(self.gitlab_id, "pagination")
        except Exception:
            pass

        self.order_by = None
        try:
            self.order_by = self._config.get(self.gitlab_id, "order_by")
        except Exception:
            pass



#!/usr/bin/env python

from urllib.parse import urljoin
from requests_html import HTMLSession

ENDPOINT = "http://localhost:8080"
LOGIN = "root"
PASSWORD = "5iveL!fe"


class GitlabSession(HTMLSession):
    def __init__(self, endpoint, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.endpoint = endpoint
        self.csrf = None

    def find_csrf_token(self, html):
        param = html.find("meta[name=csrf-param]")[0].attrs["content"]
        token = html.find("meta[name=csrf-token]")[0].attrs["content"]
        self.csrf = {param: token}

    def obtain_csrf_token(self):
        r = self.get(urljoin(self.endpoint, "/"))
        self.find_csrf_token(r.html)

    def sign_in(self, login, password):
        data = {"user[login]": login, "user[password]": password, **self.csrf}
        r = self.post(urljoin(self.endpoint, "/users/sign_in"), data=data)
        self.find_csrf_token(r.html)

    def obtain_personal_access_token(self, name):
        data = {
            "personal_access_token[name]": name,
            "personal_access_token[scopes][]": ["api", "sudo"],
            **self.csrf,
        }
        r = self.post(
            urljoin(self.endpoint, "/profile/personal_access_tokens"), data=data
        )
        return r.html.find("#created-personal-access-token")[0].attrs["value"]


def main():
    with GitlabSession(ENDPOINT) as s:
        s.obtain_csrf_token()
        s.sign_in(LOGIN, PASSWORD)
        print(s.obtain_personal_access_token("default"))


if __name__ == "__main__":
    main()



"""
GitLab API: https://docs.gitlab.com/ce/api/services.html
"""

import pytest
import responses

from gitlab.v4.objects import ProjectService


@pytest.fixture
def resp_service():
    content = {
        "id": 100152,
        "title": "Pipelines emails",
        "slug": "pipelines-email",
        "created_at": "2019-01-14T08:46:43.637+01:00",
        "updated_at": "2019-07-01T14:10:36.156+02:00",
        "active": True,
        "commit_events": True,
        "push_events": True,
        "issues_events": True,
        "confidential_issues_events": True,
        "merge_requests_events": True,
        "tag_push_events": True,
        "note_events": True,
        "confidential_note_events": True,
        "pipeline_events": True,
        "wiki_page_events": True,
        "job_events": True,
        "comment_on_event_enabled": True,
        "project_id": 1,
    }

    with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/services",
            json=[content],
            content_type="application/json",
            status=200,
        )
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/services",
            json=content,
            content_type="application/json",
            status=200,
        )
        rsps.add(
            method=responses.GET,
            url="http://localhost/api/v4/projects/1/services/pipelines-email",
            json=content,
            content_type="application/json",
            status=200,
        )
        updated_content = dict(content)
        updated_content["issues_events"] = False
        rsps.add(
            method=responses.PUT,
            url="http://localhost/api/v4/projects/1/services/pipelines-email",
            json=updated_content,
            content_type="application/json",
            status=200,
        )
        yield rsps


def test_list_active_services(project, resp_service):
    services = project.services.list()
    assert isinstance(services, list)
    assert isinstance(services[0], ProjectService)
    assert services[0].active
    assert services[0].push_events


def test_list_available_services(project, resp_service):
    services = project.services.available()
    assert isinstance(services, list)
    assert isinstance(services[0], str)


def test_get_service(project, resp_service):
    service = project.services.get("pipelines-email")
    assert isinstance(service, ProjectService)
    assert service.push_events is True


def test_update_service(project, resp_service):
    service = project.services.get("pipelines-email")
    service.issues_events = False
    service.save()
    assert service.issues_events is False



def test_list_project_packages(gitlab_cli, project):
    cmd = ["project-package", "list", "--project-id", project.id]
    ret = gitlab_cli(cmd)

    assert ret.success


def test_list_group_packages(gitlab_cli, group):
    cmd = ["group-package", "list", "--group-id", group.id]
    ret = gitlab_cli(cmd)

    assert ret.success
